# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
AMF0 implementation.

C{AMF0} supports the basic data types used for the NetConnection, NetStream,
LocalConnection, SharedObjects and other classes in the Adobe Flash Player.

@since: 0.1
@see: U{Official AMF0 Specification in English (external)
    <http://opensource.adobe.com/wiki/download/attachments/1114283/amf0_spec_121207.pdf>}
@see: U{Official AMF0 Specification in Japanese (external)
    <http://opensource.adobe.com/wiki/download/attachments/1114283/JP_amf0_spec_121207.pdf>}
@see: U{AMF documentation on OSFlash (external)
    <http://osflash.org/documentation/amf>}
"""

import datetime

import miniamf
from . import util, codec, xml


#: Represented as 9 bytes: 1 byte for C{0x00} and 8 bytes a double
#: representing the value of the number.
TYPE_NUMBER = b'\x00'
#: Represented as 2 bytes: 1 byte for C{0x01} and a second, C{0x00}
#: for C{False}, C{0x01} for C{True}.
TYPE_BOOL = b'\x01'
#: Represented as 3 bytes + len(String): 1 byte C{0x02}, then a UTF8 string,
#: including the top two bytes representing string length as a C{int}.
TYPE_STRING = b'\x02'
#: Represented as 1 byte, C{0x03}, then pairs of UTF8 string, the key, and
#: an AMF element, ended by three bytes, C{0x00} C{0x00} C{0x09}.
TYPE_OBJECT = b'\x03'
#: MovieClip does not seem to be supported by Remoting.
#: It may be used by other AMF clients such as SharedObjects.
TYPE_MOVIECLIP = b'\x04'
#: 1 single byte, C{0x05} indicates null.
TYPE_NULL = b'\x05'
#: 1 single byte, C{0x06} indicates null.
TYPE_UNDEFINED = b'\x06'
#: When an ActionScript object refers to itself, such C{this.self = this},
#: or when objects are repeated within the same scope (for example, as the
#: two parameters of the same function called), a code of C{0x07} and an
#: C{int}, the reference number, are written.
TYPE_REFERENCE = b'\x07'
#: A MixedArray is indicated by code C{0x08}, then a Long representing the
#: highest numeric index in the array, or 0 if there are none or they are
#: all negative. After that follow the elements in key : value pairs.
TYPE_MIXEDARRAY = b'\x08'
#: @see: L{TYPE_OBJECT}
TYPE_OBJECTTERM = b'\x09'
#: An array is indicated by C{0x0A}, then a Long for array length, then the
#: array elements themselves. Arrays are always sparse; values for
#: inexistant keys are set to null (C{0x06}) to maintain sparsity.
TYPE_ARRAY = b'\x0A'
#: Date is represented as C{0x0B}, then a double, then an C{int}. The double
#: represents the number of milliseconds since 01/01/1970. The C{int}
#: represents the timezone offset in minutes between GMT. Note for the latter
#: than values greater than 720 (12 hours) are represented as M{2^16} - the
#: value. Thus GMT+1 is 60 while GMT-5 is 65236.
TYPE_DATE = b'\x0B'
#: LongString is reserved for strings larger then M{2^16} characters long. It
#: is represented as C{0x0C} then a LongUTF.
TYPE_LONGSTRING = b'\x0C'
#: Trying to send values which don't make sense, such as prototypes, functions,
#: built-in objects, etc. will be indicated by a single C{00x0D} byte.
TYPE_UNSUPPORTED = b'\x0D'
#: Remoting Server -> Client only.
#: @see: L{RecordSet}
#: @see: U{RecordSet structure on OSFlash
#: <http://osflash.org/documentation/amf/recordset>}
TYPE_RECORDSET = b'\x0E'
#: The XML element is indicated by C{0x0F} and followed by a LongUTF containing
#: the string representation of the XML object. The receiving gateway may which
#: to wrap this string inside a language-specific standard XML object, or
#: simply pass as a string.
TYPE_XML = b'\x0F'
#: A typed object is indicated by C{0x10}, then a UTF string indicating class
#: name, and then the same structure as a normal C{0x03} Object. The receiving
#: gateway may use a mapping scheme, or send back as a vanilla object or
#: associative array.
TYPE_TYPEDOBJECT = b'\x10'
#: An AMF message sent from an AVM+ client such as the Flash Player 9 may break
#: out into L{AMF3<miniamf.amf3>} mode. In this case the next byte will be the
#: AMF3 type code and the data will be in AMF3 format until the decoded object
#: reaches it's logical conclusion (for example, an object has no more keys).
TYPE_AMF3 = b'\x11'


class Context(codec.Context):
    """
    """

    def clear(self):
        codec.Context.clear(self)

        encoder = self.extra.get('amf3_encoder', None)

        if encoder:
            encoder.context.clear()

        decoder = self.extra.get('amf3_decoder', None)

        if decoder:
            decoder.context.clear()

    def getAMF3Encoder(self, amf0_encoder):
        encoder = self.extra.get('amf3_encoder', None)

        if encoder:
            return encoder

        encoder = miniamf.get_encoder(
            miniamf.AMF3,
            stream=amf0_encoder.stream,
            timezone_offset=amf0_encoder.timezone_offset
        )

        self.extra['amf3_encoder'] = encoder

        return encoder

    def getAMF3Decoder(self, amf0_decoder):
        decoder = self.extra.get('amf3_decoder', None)

        if decoder:
            return decoder

        decoder = miniamf.get_decoder(
            miniamf.AMF3,
            stream=amf0_decoder.stream,
            timezone_offset=amf0_decoder.timezone_offset
        )

        self.extra['amf3_decoder'] = decoder

        return decoder


class Decoder(codec.Decoder):
    """
    Decodes an AMF0 stream.
    """

    def buildContext(self, **kwargs):
        return Context(**kwargs)

    def getTypeFunc(self, data):
        # great for coverage, sucks for readability
        if data == TYPE_NUMBER:
            return self.readNumber
        if data == TYPE_BOOL:
            return self.readBoolean
        if data == TYPE_STRING:
            return self.readString
        if data == TYPE_OBJECT:
            return self.readObject
        if data == TYPE_NULL:
            return self.readNull
        if data == TYPE_UNDEFINED:
            return self.readUndefined
        if data == TYPE_REFERENCE:
            return self.readReference
        if data == TYPE_MIXEDARRAY:
            return self.readMixedArray
        if data == TYPE_ARRAY:
            return self.readList
        if data == TYPE_DATE:
            return self.readDate
        if data == TYPE_LONGSTRING:
            return self.readLongString
        if data == TYPE_UNSUPPORTED:
            return self.readNull
        if data == TYPE_XML:
            return self.readXML
        if data == TYPE_TYPEDOBJECT:
            return self.readTypedObject
        if data == TYPE_AMF3:
            return self.readAMF3

    def readNumber(self):
        """
        Reads a ActionScript C{Number} value.

        In ActionScript 1 and 2 the C{NumberASTypes} type represents all
        numbers, both floats and integers.

        @rtype: C{int} or C{float}
        """
        return _check_for_int(self.stream.read_double())

    def readBoolean(self):
        """
        Reads a ActionScript C{Boolean} value.

        @rtype: C{bool}
        @return: Boolean.
        """
        return bool(self.stream.read_uchar())

    def readString(self, bytes=False):
        """
        Reads a C{string} from the stream. If bytes is C{True} then you will
        get the raw data read from the stream, otherwise a string that has been
        B{utf-8} decoded.
        """
        b = self.stream.read(self.stream.read_ushort())

        if bytes:
            return b

        return self.context.getStringForBytes(b)

    def readNull(self):
        """
        Reads a ActionScript C{null} value.
        """
        return None

    def readUndefined(self):
        """
        Reads an ActionScript C{undefined} value.

        @return: L{Undefined<miniamf.Undefined>}
        """
        return miniamf.Undefined

    def readMixedArray(self):
        """
        Read mixed array.

        @rtype: L{miniamf.MixedArray}
        """
        # TODO: something with the length/strict
        self.stream.read_ulong()  # length

        obj = miniamf.MixedArray()
        self.context.addObject(obj)

        attrs = self.readObjectAttributes(obj)

        for key, value in attrs.items():
            try:
                key = int(key)
            except ValueError:
                pass

            obj[key] = value

        return obj

    def readList(self):
        """
        Read a C{list} from the data stream.
        """
        obj = []
        self.context.addObject(obj)
        l = self.stream.read_ulong()

        for i in range(l):
            obj.append(self.readElement())

        return obj

    def readTypedObject(self):
        """
        Reads an aliased ActionScript object from the stream and attempts to
        'cast' it into a python class.

        @see: L{miniamf.register_class}
        """
        class_alias = self.readString()

        try:
            alias = self.context.getClassAlias(class_alias)
        except miniamf.UnknownClassAlias:
            if self.strict:
                raise

            alias = miniamf.TypedObjectClassAlias(class_alias)

        obj = alias.createInstance(codec=self)
        self.context.addObject(obj)

        attrs = self.readObjectAttributes(obj)
        alias.applyAttributes(obj, attrs, codec=self)

        return obj

    def readAMF3(self):
        """
        Read AMF3 elements from the data stream.

        @return: The AMF3 element read from the stream
        """
        return self.context.getAMF3Decoder(self).readElement()

    def readObjectAttributes(self, obj):
        obj_attrs = {}

        key = self.readString()

        while self.stream.peek() != TYPE_OBJECTTERM:
            obj_attrs[key] = self.readElement()
            key = self.readString()

        # discard the end marker (TYPE_OBJECTTERM)
        self.stream.read(1)

        return obj_attrs

    def readObject(self):
        """
        Reads an anonymous object from the data stream.

        @rtype: L{ASObject<miniamf.ASObject>}
        """
        obj = miniamf.ASObject()
        self.context.addObject(obj)

        obj.update(self.readObjectAttributes(obj))

        return obj

    def readReference(self):
        """
        Reads a reference from the data stream.

        @raise miniamf.ReferenceError: Unknown reference.
        """
        idx = self.stream.read_ushort()
        o = self.context.getObject(idx)

        if o is None:
            raise miniamf.ReferenceError('Unknown reference %d' % (idx,))

        return o

    def readDate(self):
        """
        Reads a UTC date from the data stream. Client and servers are
        responsible for applying their own timezones.

        Date: C{0x0B T7 T6} .. C{T0 Z1 Z2 T7} to C{T0} form a 64 bit
        Big Endian number that specifies the number of nanoseconds
        that have passed since 1/1/1970 0:00 to the specified time.
        This format is UTC 1970. C{Z1} and C{Z0} for a 16 bit Big
        Endian number indicating the indicated time's timezone in
        minutes.
        """
        ms = self.stream.read_double() / 1000.0
        self.stream.read_short()  # tz

        # Timezones are ignored
        d = util.get_datetime(ms)

        if self.timezone_offset:
            d = d + self.timezone_offset

        self.context.addObject(d)

        return d

    def readLongString(self):
        """
        Read UTF8 string.
        """
        l = self.stream.read_ulong()

        bytes = self.stream.read(l)

        return self.context.getStringForBytes(bytes)

    def readXML(self):
        """
        Read XML.
        """
        root = xml.fromstring(
            self.readLongString(),
            forbid_dtd=self.context.forbid_dtd,
            forbid_entities=self.context.forbid_entities,
        )

        self.context.addObject(root)

        return root


class Encoder(codec.Encoder):
    """
    Encodes an AMF0 stream.

    @ivar use_amf3: A flag to determine whether this encoder should default to
        using AMF3. Defaults to C{False}
    @type use_amf3: C{bool}
    """

    def __init__(self, *args, **kwargs):
        codec.Encoder.__init__(self, *args, **kwargs)

        self.use_amf3 = kwargs.pop('use_amf3', False)

    def buildContext(self, **kwargs):
        return Context(**kwargs)

    def getTypeFunc(self, data):
        if self.use_amf3:
            return self.writeAMF3

        if type(data) is miniamf.MixedArray:
            return self.writeMixedArray

        return codec.Encoder.getTypeFunc(self, data)

    def writeType(self, t):
        """
        Writes the type to the stream.

        @type   t: byte string
        @param  t: ActionScript type.
        """
        self.stream.write(t)

    def writeUndefined(self, data):
        """
        Writes the L{undefined<TYPE_UNDEFINED>} data type to the stream.

        @param data: Ignored, here for the sake of interface.
        """
        self.writeType(TYPE_UNDEFINED)

    def writeNull(self, n):
        """
        Write null type to data stream.
        """
        self.writeType(TYPE_NULL)

    def writeList(self, a):
        """
        Write array to the stream.

        @param a: The array data to be encoded to the AMF0 data stream.
        """
        if self.writeReference(a) != -1:
            return

        self.context.addObject(a)

        self.writeType(TYPE_ARRAY)
        self.stream.write_ulong(len(a))

        for data in a:
            self.writeElement(data)

    def writeNumber(self, n):
        """
        Write number to the data stream .

        @param  n: The number data to be encoded to the AMF0 data stream.
        """
        self.writeType(TYPE_NUMBER)
        self.stream.write_double(float(n))

    def writeBoolean(self, b):
        """
        Write boolean to the data stream.

        @param b: The boolean data to be encoded to the AMF0 data stream.
        """
        self.writeType(TYPE_BOOL)

        if b:
            self.stream.write_uchar(1)
        else:
            self.stream.write_uchar(0)

    def serialiseString(self, s):
        """
        Similar to L{writeString} but does not encode a type byte.
        """
        if isinstance(s, str):
            s = self.context.getBytesForString(s)

        l = len(s)

        if l > 0xffff:
            self.stream.write_ulong(l)
        else:
            self.stream.write_ushort(l)

        self.stream.write(s)

    def writeBytes(self, s):
        """
        Write a string of bytes to the data stream.
        """
        l = len(s)

        if l > 0xffff:
            self.writeType(TYPE_LONGSTRING)
        else:
            self.writeType(TYPE_STRING)

        if l > 0xffff:
            self.stream.write_ulong(l)
        else:
            self.stream.write_ushort(l)

        self.stream.write(s)

    def writeString(self, u):
        """
        Write a Unicode string to the data stream.
        """
        s = self.context.getBytesForString(u)

        self.writeBytes(s)

    def writeReference(self, o):
        """
        Write reference to the data stream.

        @param o: The reference data to be encoded to the AMF0 datastream.
        """
        idx = self.context.getObjectReference(o)

        if idx == -1 or idx > 65535:
            return -1

        self.writeType(TYPE_REFERENCE)
        self.stream.write_ushort(idx)

        return idx

    def _writeDict(self, o):
        """
        Write C{dict} to the data stream.

        @param o: The C{dict} data to be encoded to the AMF0 data stream.
        """

        # For consistency with AMF3, all string indices are output first,
        # then all integer indices.  We don't do any of the other special
        # handling for integer indices required in AMF3, though.
        int_items = []
        str_items = []

        for k, v in o.items():
            if isinstance(k, int):
                int_items.append((str(k).encode("utf-8"), v))
            elif isinstance(k, str):
                str_items.append((k.encode("utf-8"), v))
            elif isinstance(k, bytes):
                str_items.append((k, v))
            else:
                raise miniamf.EncodeError(
                    "Unable to encode key %r in dict %r" % (k, o))

        int_items.sort()
        str_items.sort()

        for k, v in int_items:
            self.serialiseString(k)
            self.writeElement(v)

        for k, v in str_items:
            self.serialiseString(k)
            self.writeElement(v)

    def writeMixedArray(self, o):
        """
        Write mixed array to the data stream.

        @type o: L{miniamf.MixedArray}
        """
        if self.writeReference(o) != -1:
            return

        self.context.addObject(o)
        self.writeType(TYPE_MIXEDARRAY)

        # TODO: optimise this
        # work out the highest integer index
        try:
            # list comprehensions to save the day
            max_index = max([
                y for y in o.keys()
                if isinstance(y, int)
            ])

            if max_index < 0:
                max_index = 0
        except ValueError:
            max_index = 0

        self.stream.write_ulong(max_index)

        self._writeDict(o)
        self._writeEndObject()

    def _writeEndObject(self):
        self.stream.write(b'\x00\x00' + TYPE_OBJECTTERM)

    def writeObject(self, o):
        """
        Write a Python object to the stream.

        @param o: The object data to be encoded to the AMF0 data stream.
        """
        if self.writeReference(o) != -1:
            return

        self.context.addObject(o)
        alias = self.context.getClassAlias(o.__class__)

        alias.compile()

        if alias.amf3:
            self.writeAMF3(o)

            return

        if alias.anonymous:
            self.writeType(TYPE_OBJECT)
        else:
            self.writeType(TYPE_TYPEDOBJECT)
            self.serialiseString(alias.alias)

        attrs = alias.getEncodableAttributes(o, codec=self)

        if alias.static_attrs and attrs:
            for key in sorted(alias.static_attrs):
                value = attrs.pop(key)

                self.serialiseString(key)
                self.writeElement(value)

        if attrs:
            self._writeDict(attrs)

        self._writeEndObject()

    def writeDate(self, d):
        """
        Writes a date to the data stream.

        @type d: Instance of C{datetime.datetime}
        @param d: The date to be encoded to the AMF0 data stream.
        """
        if isinstance(d, datetime.time):
            raise miniamf.EncodeError(
                'A datetime.time instance was found but AMF0 has no way to '
                'encode time objects. Please use datetime.datetime instead '
                '(got:%r)' % (d,)
            )

        # According to the Red5 implementation of AMF0, dates references are
        # created, but not used.
        if self.timezone_offset is not None:
            d -= self.timezone_offset

        secs = util.get_timestamp(d)

        self.writeType(TYPE_DATE)
        self.stream.write_double(secs * 1000.0)
        self.stream.write_short(0)  # tz

    def writeXML(self, e):
        """
        Writes an XML instance.
        """
        self.writeType(TYPE_XML)

        data = xml.tostring(e)

        if isinstance(data, str):
            data = data.encode('utf-8')

        self.stream.write_ulong(len(data))
        self.stream.write(data)

    def writeAMF3(self, data):
        """
        Writes an element in L{AMF3<miniamf.amf3>} format.
        """
        self.writeType(TYPE_AMF3)

        self.context.getAMF3Encoder(self).writeElement(data)


class RecordSet(object):
    """
    I represent the C{RecordSet} class used in Adobe Flash Remoting to hold
    (amongst other things) SQL records.

    @ivar columns: The columns to send.
    @type columns: List of strings.
    @ivar items: The C{RecordSet} data.
    @type items: List of lists, the order of the data corresponds to the order
        of the columns.
    @ivar service: Service linked to the C{RecordSet}.
    @type service:
    @ivar id: The id of the C{RecordSet}.
    @type id: string

    @see: U{RecordSet on OSFlash (external)
        <http://osflash.org/documentation/amf/recordset>}
    """

    class __amf__:
        alias = 'RecordSet'
        static = ('serverInfo',)
        dynamic = False

    def __init__(self, columns=[], items=[], service=None, id=None):
        self.columns = columns
        self.items = items
        self.service = service
        self.id = id

    def _get_server_info(self):
        ret = miniamf.ASObject(
            totalCount=len(self.items),
            cursor=1,
            version=1,
            initialData=self.items,
            columnNames=self.columns
        )

        if self.service is not None:
            ret.update({'serviceName': str(self.service['name'])})

        if self.id is not None:
            ret.update({'id': str(self.id)})

        return ret

    def _set_server_info(self, val):
        self.columns = val['columnNames']
        self.items = val['initialData']

        try:
            # TODO nick: find relevant service and link in here.
            self.service = dict(name=val['serviceName'])
        except KeyError:
            self.service = None

        try:
            self.id = val['id']
        except KeyError:
            self.id = None

    serverInfo = property(_get_server_info, _set_server_info)

    def __repr__(self):
        ret = '<%s.%s' % (self.__module__, self.__class__.__name__)

        if self.id is not None:
            ret += ' id=%s' % self.id

        if self.service is not None:
            ret += ' service=%s' % self.service

        ret += ' at 0x%x>' % id(self)

        return ret


miniamf.register_class(RecordSet)


def _check_for_int(x):
    """
    This is a compatibility function that takes a C{float} and converts it to
    an C{int} if the values are equal.
    """
    try:
        y = int(x)
    except (OverflowError, ValueError):
        pass
    else:
        # There is no way in AMF0 to distinguish between integers and floats
        if x == x and y == x:
            return y

    return x



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
C-extension for L{miniamf.util} Python module in L{Mini-AMF<miniamf>}.

@since: 0.4
"""

from cpython cimport *
from libc.stdlib cimport *
from libc.string cimport *

cimport cython

cdef extern from "stdio.h":
    int SIZEOF_LONG

cdef extern from "Python.h":
    int PyFloat_Pack4(double, char *, int) except? -1
    int PyFloat_Pack8(double, char *, int) except? -1
    double PyFloat_Unpack4(const char *, int) except? -1.0
    double PyFloat_Unpack8(const char *, int) except? -1.0
    int PyByteArray_Check(object)
    char* PyByteArray_AsString(object)

# module constant declarations
cdef char ENDIAN_NETWORK = "!"
cdef char ENDIAN_NATIVE = "@"
cdef char ENDIAN_LITTLE = "<"
cdef char ENDIAN_BIG = ">"

from miniamf.util.pure import ENDIAN_NETWORK as PyEndianNetwork
from miniamf.util.pure import ENDIAN_NATIVE as PyEndianNative
from miniamf.util.pure import ENDIAN_LITTLE as PyEndianLittle
from miniamf.util.pure import ENDIAN_BIG as PyEndianBig

cdef int MAX_BUFFER_EXTENSION = 1 << 14

cdef char SYSTEM_ENDIAN

cdef int float_broken = -1

cdef const char *NaN = <const char *>'\xff\xf8\x00\x00\x00\x00\x00\x00'
cdef const char *NegInf = <const char *>'\xff\xf0\x00\x00\x00\x00\x00\x00'
cdef const char *PosInf = <const char *>'\x7f\xf0\x00\x00\x00\x00\x00\x00'

cdef double platform_nan
cdef double platform_posinf
cdef double platform_neginf

cdef double system_nan
cdef double system_posinf
cdef double system_neginf

cdef object miniamf_NaN = PyFloat_FromDouble(1e300000 / -1e300000)
cdef object miniamf_NegInf = PyFloat_FromDouble(-1e300000)
cdef object miniamf_PosInf = PyFloat_FromDouble(1e300000)
cdef object empty_unicode = unicode('')

@cython.profile(False)
cdef int _memcpy_ensure_endian(void *src, void *dest, unsigned int size) nogil:
    """
    """
    cdef char *buf = <char *>malloc(size)

    if buf == NULL:
        return -1

    memcpy(buf, src, size)

    if not is_big_endian(SYSTEM_ENDIAN):
        if swap_bytes(buf, size) == -1:
            free(buf)

            return -1

    memcpy(dest, buf, size)
    free(buf)

    return 0


cdef int build_platform_exceptional_floats() except -1:
    global platform_nan, platform_posinf, platform_neginf
    global system_nan, system_posinf, system_neginf

    if _memcpy_ensure_endian(<void *>NaN, &system_nan, 8) == -1:
        PyErr_NoMemory()

    if _memcpy_ensure_endian(<void *>NegInf, &system_neginf, 8) == -1:
        PyErr_NoMemory()

    if _memcpy_ensure_endian(<void *>PosInf, &system_posinf, 8) == -1:
        PyErr_NoMemory()

    if float_broken == 1:
        try:
            PyFloat_Unpack8(NaN, not is_big_endian(SYSTEM_ENDIAN))
        except:
            raise
        else:
            memcpy(&platform_nan, &system_nan, 8)

        try:
            PyFloat_Unpack8(PosInf, not is_big_endian(SYSTEM_ENDIAN))
        except:
            raise
        else:
            memcpy(&platform_posinf, &system_posinf, 8)

        try:
            PyFloat_Unpack8(NegInf, not is_big_endian(SYSTEM_ENDIAN))
        except:
            raise
        else:
            memcpy(&platform_neginf, &system_neginf, 8)

    return 0


@cython.profile(False)
cdef char get_native_endian() nogil:
    """
    A quick hack to determine the system's endian-ness ...

    @return: Either L{ENDIAN_LITTLE} or L{ENDIAN_BIG}
    """
    cdef unsigned int one = 1
    cdef int big_endian = (<char*>&one)[0] != 1

    if big_endian == 1:
        return ENDIAN_BIG

    return ENDIAN_LITTLE


@cython.profile(False)
cdef inline bint is_big_endian(char endian) nogil:
    """
    Returns a boolean value whether the supplied C{endian} is big.
    """
    if endian == ENDIAN_NATIVE:
        return SYSTEM_ENDIAN == ENDIAN_BIG

    return endian == ENDIAN_NETWORK or endian == ENDIAN_BIG


@cython.profile(False)
cdef inline int is_native_endian(char endian) nogil:
    if endian == ENDIAN_NATIVE:
        return 1

    if endian == ENDIAN_NETWORK:
        endian = ENDIAN_BIG

    return endian == SYSTEM_ENDIAN


@cython.profile(False)
cdef inline int swap_bytes(char *buffer, Py_ssize_t size) nogil:
    cdef char *buf = <char *>malloc(size)

    if buf == NULL:
        return -1

    cdef Py_ssize_t i

    for i from 0 <= i < size:
        buf[i] = buffer[size - i - 1]

    memcpy(buffer, buf, size)
    free(buf)

    return 0


cdef bint is_broken_float() except -1:
    cdef double test = PyFloat_Unpack8(NaN, 0)

    cdef int result
    cdef char *buf = <char *>&test

    if not is_big_endian(SYSTEM_ENDIAN):
        if swap_bytes(buf, 8) == -1:
            PyErr_NoMemory()

    result = memcmp(NaN, buf, 8)

    return result != 0


cdef class cBufferedByteStream(object):
    """
    A file like object that can be read/written to. Supports data type
    specific reads/writes (e.g. ints, longs, floats etc.)

    Endian aware.
    """

    def __cinit__(self):
        self.endian = ENDIAN_NETWORK
        self.buffer = NULL
        self.min_buf_size = 512
        self.size = 0

    def __dealloc__(self):
        if self.buffer != NULL:
            free(self.buffer)

        self.buffer = NULL

    cdef inline int _init_buffer(self) except -1:
        if self.buffer != NULL:
            free(self.buffer)

            self.buffer = NULL

        self.pos = 0
        self.length = 0
        self.size = self.min_buf_size

        self.buffer = <char *>malloc(self.size)

        if self.buffer == NULL:
            PyErr_NoMemory()

        return 0

    cpdef Py_ssize_t tell(self) except -1:
        """
        Returns the position of the stream pointer.
        """
        return self.pos

    cdef int _actually_increase_buffer(self, Py_ssize_t new_size) except -1:
        if self.size == 0:
            self._init_buffer()

        cdef Py_ssize_t requested_size = self.size
        cdef char *buf

        while new_size > requested_size:
            requested_size *= 2

        if requested_size > new_size + MAX_BUFFER_EXTENSION:
            requested_size = new_size + MAX_BUFFER_EXTENSION

        buf = <char *>realloc(self.buffer, sizeof(char *) * requested_size)

        if buf == NULL:
            PyErr_NoMemory()

        self.buffer = buf
        self.size = requested_size

        return 0

    cdef inline int _increase_buffer(self, Py_ssize_t size) except -1:
        """
        Request to increase the buffer by C{size}
        """
        cdef Py_ssize_t new_len = self.length + size

        if new_len <= self.size:
            return 0

        return self._actually_increase_buffer(new_len)

    cdef int write(self, char *buf, Py_ssize_t size) except -1:
        """
        Writes the content of the specified C{buf} into this buffer.
        """
        assert buf != NULL, 'buf cannot be NULL'

        if size == 0:
            return 0

        self._increase_buffer(size)

        memcpy(self.buffer + self.pos, buf, size)

        self.pos += size

        if self.pos > self.length:
            self.length = self.pos

        return 0

    cdef inline bint has_available(self, Py_ssize_t size) except -1:
        if size == 0:
            return 1
        elif size < 0:
            return 0

        if self.length == self.pos:
            return 0

        if self.pos + size > self.length:
            return 0

        return 1

    cdef int read(self, char **buf, Py_ssize_t size) except -1:
        """
        Reads up to the specified number of bytes from the stream into
        the specified byte array of specified length.

        Do not free the results or bad things will happen
        """
        if size == -1:
            size = self.remaining()

            if size == 0:
                size = 1

        if not self.has_available(size):
            raise IOError

        buf[0] = self.buffer + self.pos

        self.pos += size

        return 0

    cpdef bint at_eof(self) except -1:
        """
        Returns C{True} if the internal pointer is at the end of the stream.

        @rtype: C{bool}
        """
        return self.length == self.pos

    cpdef Py_ssize_t remaining(self) except -1:
        """
        Returns number of remaining bytes.
        """
        return self.length - self.pos

    cpdef int seek(self, Py_ssize_t pos, int mode=0) except -1:
        """
        Sets the file-pointer offset, measured from the beginning of this stream,
        at which the next write operation will occur.

        @param mode: mode 0: absolute; 1: relative; 2: relative to EOF
        """
        if mode == 0:
            if pos < 0 or pos > self.length:
                raise IOError

            self.pos = pos
        elif mode == 1:
            if pos + self.pos < 0 or pos + self.pos > self.length:
                raise IOError

            self.pos += pos
        elif mode == 2:
            if pos + self.length < 0 or pos + self.length > self.length:
                raise IOError

            self.pos = self.length + pos
        else:
            raise ValueError('Bad value for mode')

        return 0

    cpdef object getvalue(self):
        """
        Get raw data from buffer.
        """
        return PyBytes_FromStringAndSize(self.buffer, self.length)

    cdef Py_ssize_t peek(self, char **buf, Py_ssize_t size) except -1:
        """
        Makes a pointer reference to the underlying buffer. Do NOT modify the
        returned value or free its contents. That would be seriously bad.
        """
        if not self.has_available(size):
            size = self.length - self.pos

        buf[0] = self.buffer + self.pos

        return size

    cpdef int truncate(self, Py_ssize_t size=0) except -1:
        """
        Truncates the stream to the specified length.

        @param size: The length of the stream, in bytes.
        @type size: C{int}
        """
        if size > self.length:
            raise IOError

        if size == 0:
            return self._init_buffer()

        cdef char *buf = NULL
        cdef Py_ssize_t cur_pos = self.pos

        buf = <char *>malloc(self.length)

        if buf == NULL:
            PyErr_NoMemory()

        memcpy(buf, self.buffer, self.length)

        try:
            self._init_buffer()

            self.write(buf, size)
        finally:
            free(buf)

        if self.length > cur_pos:
            self.pos = self.length
        else:
            self.seek(cur_pos, 0)

        return 0

    cpdef int consume(self) except -1:
        """
        Chops the tail off the stream starting at 0 and ending at C{tell()}.
        The stream pointer is set to 0 at the end of this function.
        """
        cdef char *buf = NULL
        cdef char *peek_buf
        cdef Py_ssize_t size = self.remaining()

        if size > 0:
            size = self.peek(&peek_buf, size)
            buf = <char *>malloc(size)

            if buf == NULL:
                PyErr_NoMemory()

            memcpy(buf, peek_buf, size)

        try:
            self._init_buffer()

            if size > 0:
                self.write(buf, size)
        finally:
            free(buf)

        self.pos = 0

        return 0

    cdef int unpack_int(self, unsigned int num_bytes, void *ret) except -1:
        """
        Unpacks a long from C{buf}.
        """
        cdef Py_ssize_t nb

        if num_bytes > 4:
            raise ValueError('Max 4 bytes to unpack')

        if not self.has_available(num_bytes):
            raise IOError

        cdef unsigned long x = 0
        cdef unsigned char *bytes = <unsigned char *>(self.buffer + self.pos)

        if is_big_endian(self.endian):
            for 0 <= nb < num_bytes:
                x = (x << 8) | bytes[nb]
        else:
            for 0 <= nb < num_bytes:
                x = (x << 8) | bytes[num_bytes - nb - 1]

        self.pos += num_bytes
        memcpy(ret, &x, num_bytes)

        return 0

    cdef int pack_int(self, int num_bytes, long x) except -1:
        """
        Packs a long.

        @raise OverflowError: integer out of range
        """
        cdef long maxint = 1
        cdef long minint = -1

        if num_bytes != SIZEOF_LONG:
            maxint = (maxint << (num_bytes * 8 - 1)) - 1
            minint = (-maxint) - 1

            if x > maxint or x < minint:
                raise OverflowError('integer out of range')

        cdef char *buf = <char *>malloc(num_bytes)

        if buf == NULL:
            PyErr_NoMemory()

        cdef long i = num_bytes

        if is_big_endian(self.endian):
            while i > 0:
                i -= 1
                buf[i] = <char>x
                x >>= 8
        else:
            while i > 0:
                buf[num_bytes - i] = <char>x
                i -= 1
                x >>= 8

        try:
            self.write(buf, num_bytes)
        finally:
            free(buf)

        return 0

    cdef int pack_uint(self, int num_bytes, unsigned long x) except -1:
        """
        Packs an unsigned long into a buffer.

        @raise OverflowError: integer out of range
        """
        cdef unsigned long maxint = 1

        if num_bytes != SIZEOF_LONG:
            maxint <<= <unsigned long>(num_bytes * 8)

            if x >= maxint:
                raise OverflowError('integer out of range')

        cdef char *buf = <char *>malloc(num_bytes)

        if buf == NULL:
            PyErr_NoMemory()

        cdef long i = num_bytes

        if is_big_endian(self.endian):
            while i > 0:
                i -= 1
                buf[i] = <char>x
                x >>= 8
        else:
            while i > 0:
                buf[num_bytes - i] = <char>x
                i -= 1
                x >>= 8

        try:
            self.write(buf, num_bytes)
        finally:
            free(buf)

        return 0

    cpdef unsigned char read_uchar(self) except? 0:
        """
        Reads an C{unsigned char} from the stream.
        """
        cdef unsigned char ch = 0

        self.unpack_int(1, &ch)

        return ch

    cpdef char read_char(self) except? 0:
        """
        Reads a C{char} from the stream.
        """
        cdef char ch = 0

        self.unpack_int(1, &ch)

        return ch

    cpdef unsigned short read_ushort(self) except? 0:
        """
        Reads a 2 byte unsigned integer from the stream.
        """
        cdef unsigned short x = 0

        self.unpack_int(2, &x)

        return x

    cpdef short read_short(self) except? 0:
        """
        Reads a 2 byte integer from the stream.
        """
        cdef short x = 0

        self.unpack_int(2, &x)

        return x

    cpdef unsigned long read_24bit_uint(self) except? 0:
        """
        Reads a 24 bit unsigned integer from the stream.
        """
        cdef unsigned long x = 0

        self.unpack_int(3, &x)

        return x

    cpdef long read_24bit_int(self) except? 0:
        """
        Reads a 24 bit integer from the stream.
        """
        cdef long x = 0

        self.unpack_int(3, &x)

        if x & 0x800000:
            x |= ~0xffffff

        return x

    cpdef unsigned long read_ulong(self) except? 0:
        """
        Reads a 4 byte unsigned integer from the stream.
        """
        cdef unsigned long x = 0

        self.unpack_int(4, &x)

        return x

    cpdef long read_long(self) except? 0:
        """
        Reads a 4 byte integer from the stream.
        """
        cdef long x = 0

        self.unpack_int(4, &x)

        if x & 0x80000000:
            x |= ~0x7fffffff

        return x

    cpdef int write_uchar(self, unsigned char ret) except -1:
        """
        Writes an C{unsigned char} to the stream.

        @param ret: Unsigned char
        @type ret: C{int}
        """
        return self.pack_uint(1, <unsigned long>ret)

    cpdef int write_char(self, char ret) except -1:
        """
        Write a C{char} to the stream.

        @param ret: char
        @type ret: C{int}
        """
        return self.pack_int(1, <long>ret)

    cpdef int write_ushort(self, unsigned short ret) except -1:
        """
        Writes a 2 byte unsigned integer to the stream.

        @param ret: 2 byte unsigned integer
        @type ret: C{int}
        """
        return self.pack_uint(2, <unsigned long>ret)

    cpdef int write_short(self, short ret) except -1:
        """
        Writes a 2 byte integer to the stream.

        @param ret: 2 byte integer
        @type ret: C{int}
        """
        return self.pack_int(2, <long>ret)

    cpdef int write_ulong(self, unsigned long ret) except -1:
        """
        Writes a 4 byte unsigned integer to the stream.

        @param ret: 4 byte unsigned integer
        @type ret: C{int}
        """
        return self.pack_uint(4, ret)

    cpdef int write_long(self, long ret) except -1:
        """
        Writes a 4 byte integer to the stream.

        @param ret: 4 byte integer
        @type ret: C{int}
        """
        return self.pack_int(4, ret)

    cpdef int write_24bit_uint(self, unsigned long ret) except -1:
        """
        Writes a 24 bit unsigned integer to the stream.

        @param ret: 24 bit unsigned integer
        @type ret: C{int}
        """
        return self.pack_uint(3, ret)

    cpdef int write_24bit_int(self, long ret) except -1:
        """
        Writes a 24 bit integer to the stream.

        @param ret: 24 bit integer
        @type ret: C{int}
        """
        return self.pack_int(3, ret)

    cpdef object read_utf8_string(self, Py_ssize_t l):
        """
        Reads a UTF-8 string from the stream.

        @rtype: C{unicode}
        """
        cdef char* buf = NULL
        cdef object ret

        if l == 0:
            return empty_unicode

        self.read(&buf, l)
        ret = PyUnicode_DecodeUTF8(buf, l, 'strict')

        return ret

    cpdef int write_utf8_string(self, object obj) except -1:
        """
        Writes a unicode object to the stream in UTF-8.

        @param obj: unicode object
        @raise TypeError: Unexpected type for str C{u}.
        """
        cdef object encoded_string = obj
        cdef char *buf = NULL
        cdef Py_ssize_t l = -1

        if PyUnicode_Check(obj):
            encoded_string = PyUnicode_AsUTF8String(obj)
        elif PyByteArray_Check(obj):
            encoded_string = PyByteArray_AsString(obj)
        elif not PyBytes_Check(obj):
            raise TypeError('value must be Unicode or str')

        PyBytes_AsStringAndSize(encoded_string, &buf, &l)
        self.write(buf, l)

        return 0

    cdef int read_double(self, double *obj) except -1:
        """
        Reads an 8 byte float from the stream.
        """
        cdef char *buf = NULL
        cdef int done = 0

        self.read(<char **>&buf, 8)

        if float_broken == 1:
            if is_big_endian(SYSTEM_ENDIAN):
                if not is_big_endian(self.endian):
                    if swap_bytes(buf, 8) == -1:
                        PyErr_NoMemory()
            else:
                if is_big_endian(self.endian):
                    if swap_bytes(buf, 8) == -1:
                        PyErr_NoMemory()

            if memcmp(buf, &system_nan, 8) == 0:
                memcpy(obj, &system_nan, 8)

                done = 1
            elif memcmp(buf, &system_posinf, 8) == 0:
                memcpy(obj, &system_posinf, 8)

                done = 1
            elif memcmp(buf, &system_neginf, 8) == 0:
                memcpy(obj, &system_neginf, 8)

                done = 1

            if done == 1:
                return 0

            if is_big_endian(SYSTEM_ENDIAN):
                if not is_big_endian(self.endian):
                    if swap_bytes(buf, 8) == -1:
                        PyErr_NoMemory()
            else:
                if is_big_endian(self.endian):
                    if swap_bytes(buf, 8) == -1:
                        PyErr_NoMemory()

        obj[0] = PyFloat_Unpack8(buf, not is_big_endian(self.endian))

        return 0

    cpdef int write_double(self, double val) except -1:
        """
        Writes an 8 byte float to the stream.

        @param val: 8 byte float
        @type val: C{float}
        """
        cdef char *buf
        cdef int done = 0

        buf = <char *>malloc(sizeof(double))

        if buf == NULL:
            PyErr_NoMemory()

        try:
            if float_broken == 1:
                if memcmp(&val, &system_nan, 8) == 0:
                    memcpy(buf, &val, 8)

                    done = 1
                elif memcmp(&val, &system_posinf, 8) == 0:
                    memcpy(buf, &val, 8)

                    done = 1
                elif memcmp(&val, &system_neginf, 8) == 0:
                    memcpy(buf, &val, 8)

                    done = 1

                if done == 1:
                    if is_big_endian(SYSTEM_ENDIAN):
                        if not is_big_endian(self.endian):
                            if swap_bytes(buf, 8) == -1:
                                PyErr_NoMemory()
                    else:
                        if is_big_endian(self.endian):
                            if swap_bytes(buf, 8) == -1:
                                PyErr_NoMemory()

            if done == 0:
                PyFloat_Pack8(val, buf, not is_big_endian(self.endian))

            self.write(buf, 8)
        finally:
            free(buf)

        return 0

    cdef int read_float(self, float *x) except -1:
        """
        Reads a 4 byte float from the stream.
        """
        cdef char *buf = NULL

        self.read(&buf, 4)

        x[0] = PyFloat_Unpack4(<const char *>buf, not is_big_endian(self.endian))

        return 0

    cpdef int write_float(self, float c) except -1:
        """
        Writes a 4 byte float to the stream.

        @param c: 4 byte float
        @type c: C{float}
        """
        cdef char *buf = <char *>malloc(4)

        if buf == NULL:
            PyErr_NoMemory()

        try:
            PyFloat_Pack4(c, buf, not is_big_endian(self.endian))

            self.write(buf, 4)
        finally:
            free(buf)

        return 0

    cpdef int append(self, object obj) except -1:
        cdef int i = self.pos

        self.pos = self.length

        if hasattr(obj, 'getvalue'):
            self.write_utf8_string(obj.getvalue())
        else:
            self.write_utf8_string(obj)

        self.pos = i

        return 0

    def __bool__(self):
        return self.length > 0


cdef class BufferedByteStream(cBufferedByteStream):
    """
    A Python exposed version of cBufferedByteStream. This exists because of
    various intricacies of Cythons cpdef (probably just user stupidity tho)
    """

    def __init__(self, buf=None, min_buf_size=512):
        cdef Py_ssize_t i
        cdef cBufferedByteStream x

        self.min_buf_size = min_buf_size

        if buf is None:
            pass
        elif isinstance(buf, cBufferedByteStream):
            x = <cBufferedByteStream>buf
            self.write(x.getvalue())
        elif isinstance(buf, (str, unicode)):
            self.write(PyBytes_FromString(buf))
        elif isinstance(buf, bytes):
            self.write(buf)
        elif hasattr(buf, 'getvalue'):
            self.write(buf.getvalue())
        elif hasattr(buf, 'read') and hasattr(buf, 'seek') and hasattr(buf, 'tell'):
            old_pos = buf.tell()
            buf.seek(0)
            self.write(buf.read())
            buf.seek(old_pos)
        else:
            raise TypeError("Unable to coerce buf->StringIO")

        self.seek(0)

    property endian:
        def __set__(self, value):
            if PyUnicode_Check(value) == 0:
                raise TypeError('String value expected')

            if value not in [PyEndianNetwork, PyEndianNative, PyEndianLittle, PyEndianBig]:
                raise ValueError('Not a valid endian type')

            self.endian = PyUnicode_AsUTF8String(value)[0]

        def __get__(self):
            return PyUnicode_FromStringAndSize(&self.endian, 1)

    def read(self, size=-1):
        """
        Reads C{size} bytes from the stream.
        """
        cdef Py_ssize_t s

        if size != -1:
            s = <Py_ssize_t>size
        else:
            s = self.remaining()

            if s == 0:
                s = 1

        cdef char *buf = NULL

        cBufferedByteStream.read(self, &buf, s)

        return PyBytes_FromStringAndSize(buf, s)

    def write(self, x, size=-1):
        """
        Writes the content of the specified C{x} into this buffer.

        @param x:
        @type x:
        """
        if size == -1:
            cBufferedByteStream.write_utf8_string(self, x)
        else:
            cBufferedByteStream.write(self, x, size)

    def flush(self):
        # no-op
        pass

    def __len__(self):
        return self.length

    def peek(self, Py_ssize_t size=1):
        """
        Looks C{size} bytes ahead in the stream, returning what it finds,
        returning the stream pointer to its initial position.

        @param size: Default is 1.
        @type size: C{int}

        @rtype:
        @return: Bytes.
        """
        cdef char *buf = NULL

        if size == -1:
            size = cBufferedByteStream.remaining(self)

        size = cBufferedByteStream.peek(self, &buf, size)

        return PyBytes_FromStringAndSize(buf, size)

    def write_char(self, x):
        """
        Write a C{char} to the stream.

        @param x: char
        @type x: C{int}
        @raise TypeError: Unexpected type for int C{x}.
        """
        if PyLong_Check(x) == 0:
            raise TypeError('expected int for x')

        cBufferedByteStream.write_char(self, <char>x)

    def write_ushort(self, x):
        """
        Writes a 2 byte unsigned integer to the stream.

        @param x: 2 byte unsigned integer
        @type x: C{int}
        @raise TypeError: Unexpected type for int C{x}.
        """
        if PyLong_Check(x) == 0:
            raise TypeError('expected int for x')

        cBufferedByteStream.write_ushort(self, <unsigned short>x)

    def write_short(self, x):
        """
        Writes a 2 byte integer to the stream.

        @param x: 2 byte integer
        @type x: C{int}
        @raise TypeError: Unexpected type for int C{x}.
        """
        if PyLong_Check(x) == 0:
            raise TypeError('expected int for x')

        cBufferedByteStream.write_short(self, <short>x)

    def write_ulong(self, x):
        """
        Writes a 4 byte unsigned integer to the stream.

        @param x: 4 byte unsigned integer
        @type x: C{int}
        @raise TypeError: Unexpected type for int C{x}.
        """
        if PyLong_Check(x) == 0:
            raise TypeError('expected int for x')

        if x > 4294967295L or x < 0:
            raise OverflowError

        cBufferedByteStream.write_ulong(self, <unsigned long>x)

    def read_double(self):
        """
        Reads an 8 byte float from the stream.
        """
        cdef double x

        cBufferedByteStream.read_double(self, &x)

        if float_broken == 1:
            if memcmp(&x, &system_nan, 8) == 0:
                return miniamf_NaN
            elif memcmp(&x, &system_neginf, 8) == 0:
                return miniamf_NegInf
            elif memcmp(&x, &system_posinf, 8) == 0:
                return miniamf_PosInf

        return PyFloat_FromDouble(x)

    def write_double(self, val):
        """
        Writes an 8 byte float to the stream.

        @param val: 8 byte float
        @type val: C{float}
        @raise TypeError: Unexpected type for float C{val}.
        """
        if PyFloat_Check(val) == 0:
            raise TypeError('Expecting float for val')

        cdef double d = val
        cdef int done = 0

        if float_broken == 1:
            if memcmp(&d, &platform_nan, 8) == 0:
                done = 1
            elif memcmp(&d, &platform_neginf, 8) == 0:
                done = 1
            elif memcmp(&d, &platform_posinf, 8) == 0:
                done = 1

            if done == 1:
                if is_big_endian(SYSTEM_ENDIAN):
                    if not is_big_endian(self.endian):
                        if swap_bytes(<char *>&d, 8) == -1:
                            PyErr_NoMemory()
                else:
                    if is_big_endian(self.endian):
                        if swap_bytes(<char *>&d, 8) == -1:
                            PyErr_NoMemory()

                cBufferedByteStream.write(self, <char *>&d, 8)

                return

        cBufferedByteStream.write_double(self, d)

    def read_float(self):
        """
        Reads a 4 byte float from the stream.
        """
        cdef float x

        cBufferedByteStream.read_float(self, &x)

        return PyFloat_FromDouble(x)

    def __add__(self, other):
        cdef Py_ssize_t old_pos = self.tell()
        cdef Py_ssize_t old_other_pos = other.tell()

        new = BufferedByteStream(self)

        other.seek(0)
        new.seek(0, 2)
        new.write(other.read())

        self.seek(old_pos)
        other.seek(old_other_pos)
        new.seek(0)

        return new

    def __str__(self):
        return self.getvalue()


# init the module from here

SYSTEM_ENDIAN = get_native_endian()

if is_broken_float():
    float_broken = 1

if build_platform_exceptional_floats() == -1:
    raise SystemError('Unable to initialise cminiamf.util')



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
C{weakref} support.

@since: 0.6.2
"""

import weakref

import miniamf
from miniamf.adapters import util


def get_referent(reference, **kwargs):
    return reference()


miniamf.add_type(weakref.ref, get_referent)
miniamf.add_type(weakref.WeakValueDictionary, util.to_dict)
miniamf.add_type(weakref.WeakSet, util.to_list)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tools for doing dynamic imports.

@since: 0.3
"""

import sys
from importlib.machinery import ModuleSpec

__all__ = ['when_imported']


def when_imported(name, *hooks):
    """
    Call C{hook(module)} when module named C{name} is first imported. C{name}
    must be a fully qualified (i.e. absolute) module name.

    C{hook} must accept one argument: which will be the imported module object.

    If the module has already been imported, 'hook(module)' is called
    immediately, and the module object is returned from this function. If the
    module has not been imported, then the hook is called when the module is
    first imported.
    """
    finder.when_imported(name, *hooks)


class ModuleFinder(object):
    """
    This is a special module finder object that executes a collection of
    callables when a specific module has been imported. An instance of this
    is placed in C{sys.meta_path}, which is consulted before C{sys.modules} -
    allowing us to provide this functionality.

    @ivar post_load_hooks: C{dict} of C{full module path -> callable} to be
        executed when the module is imported.
    @ivar loaded_modules: C{list} of modules that this finder has seen. Used
        to stop recursive imports in L{load_module}
    @see: L{when_imported}
    @since: 0.5
    """

    def __init__(self):
        self.post_load_hooks = {}
        self.loaded_modules = []

    def find_spec(self, fullname, path, target=None):
        """
        Called when an import is made. If there are hooks waiting for this
        module to be imported then we stop the normal import process and
        manually load the module.

        @param fullname: The name of the module being imported.
        @param path: The root path of the module (if a package). Ignored.
        @param target: Ignored.
        @return: If we want to hook this module, we return a C{ModuleSpec}.
            If not we return C{None} to allow the standard import process to
            continue.
        """
        if fullname in self.loaded_modules:
            return None

        if self.post_load_hooks.get(fullname, None):
            return ModuleSpec(fullname, self)

    def create_module(self, spec):
        name = spec.name
        self.loaded_modules.append(name)

        try:
            __import__(name, {}, {}, [])
        except Exception as e:
            self.loaded_modules.pop()

            raise e

        return sys.modules[name]

    def exec_module(self, module):
        name = module.__spec__.name
        try:
            self._run_hooks(name, sys.modules[name])
        except Exception as e:
            self.loaded_modules.pop()

            raise e

    def when_imported(self, name, *hooks):
        """
        @see: L{when_imported}
        """
        if name in sys.modules:
            for hook in hooks:
                hook(sys.modules[name])

            return

        h = self.post_load_hooks.setdefault(name, [])
        h.extend(hooks)

    def _run_hooks(self, name, module):
        """
        Run all hooks for a module.
        """
        for hook in self.post_load_hooks.pop(name, []):
            hook(module)

    def __getstate__(self):
        return (self.post_load_hooks.copy(), self.loaded_modules.copy())

    def __setstate__(self, state):
        self.post_load_hooks, self.loaded_modules = state


def _init():
    """
    Internal function to install the module finder.
    """
    global finder

    if finder is None:
        finder = ModuleFinder()

    if finder not in sys.meta_path:
        sys.meta_path.insert(0, finder)


finder = None
_init()



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
C-extension for L{miniamf.amf3} Python module in L{Mini-AMF<miniamf>}.

:since: 0.6
"""

from cpython cimport *
from libc.stdlib cimport *
from libc.string cimport *

cdef extern from "math.h":
    float floor(float)

from miniamf._accel cimport codec, amf3

import miniamf
from miniamf import xml, util


cdef char TYPE_NUMBER      = '\x00'
cdef char TYPE_BOOL        = '\x01'
cdef char TYPE_STRING      = '\x02'
cdef char TYPE_OBJECT      = '\x03'
cdef char TYPE_MOVIECLIP   = '\x04'
cdef char TYPE_NULL        = '\x05'
cdef char TYPE_UNDEFINED   = '\x06'
cdef char TYPE_REFERENCE   = '\x07'
cdef char TYPE_MIXEDARRAY  = '\x08'
cdef char TYPE_OBJECTTERM  = '\x09'
cdef char TYPE_ARRAY       = '\x0A'
cdef char TYPE_DATE        = '\x0B'
cdef char TYPE_LONGSTRING  = '\x0C'
cdef char TYPE_UNSUPPORTED = '\x0D'
cdef char TYPE_RECORDSET   = '\x0E'
cdef char TYPE_XML         = '\x0F'
cdef char TYPE_TYPEDOBJECT = '\x10'
cdef char TYPE_AMF3        = '\x11'


cdef object ASObject = miniamf.ASObject
cdef object UnknownClassAlias = miniamf.UnknownClassAlias


cdef class Context(codec.Context):
    cpdef int clear(self) except -1:
        codec.Context.clear(self)

        if self.amf3_context:
            self.amf3_context.clear()

        return 0


cdef class Decoder(codec.Decoder):
    """
    """

    def __cinit__(self):
        self.use_amf3 = 0

    def __init__(self, *args, **kwargs):
        self.use_amf3 = kwargs.pop('use_amf3', 0)
        self.context = kwargs.pop('context', None)

        if self.context is None:
            self.context = Context(**kwargs)

        codec.Codec.__init__(self, *args, **kwargs)

    cdef object readNumber(self):
        cdef double i = -1

        self.stream.read_double(&i)

        if floor(i) == i:
            try:
                return int(i)
            except OverflowError:
                return i

        return i

    cdef object readBoolean(self):
        cdef unsigned char b = self.stream.read_uchar()

        if b == 1:
            return True
        elif b == 0:
            return False

        raise miniamf.DecodeError('Bad boolean read from stream')

    cdef object readBytes(self):
        return self.context.getBytesForString(self.readString())

    cpdef object readString(self):
        cdef unsigned short l
        cdef char *b = NULL

        l = self.stream.read_ushort()

        self.stream.read(&b, l)

        return PyUnicode_DecodeUTF8(b, <Py_ssize_t>l, 'strict')

    cdef void readObjectAttributes(self, object obj_attrs):
        cdef char *peek = NULL

        while True:
            self.stream.peek(&peek, 3)

            if memcmp(peek, b'\x00\x00\x09', 3) == 0:
                self.stream.seek(3, 1)

                break

            PyDict_SetItem(obj_attrs, self.readString(), self.readElement())

        # discard the end marker (TYPE_OBJECTTERM)

    cdef object readObject(self):
        cdef object obj = ASObject()

        self.context.addObject(obj)

        self.readObjectAttributes(obj)

        return obj

    cdef object readTypedObject(self):
        cdef object class_alias = self.readString()

        try:
            alias = self.context.getClassAlias(class_alias)
        except UnknownClassAlias:
            if self.strict:
                raise

            alias = miniamf.TypedObjectClassAlias(class_alias)

        obj = alias.createInstance(codec=self)
        self.context.addObject(obj)

        cdef dict attrs = {}

        self.readObjectAttributes(attrs)

        alias.applyAttributes(obj, attrs, codec=self)

        return obj

    cdef object readReference(self):
        cdef unsigned short idx

        idx = self.stream.read_ushort()
        o = self.context.getObject(idx)

        if o is None:
            raise miniamf.ReferenceError('Unknown reference %d' % (idx,))

        return o

    cdef object readMixedArray(self):
        cdef dict attrs = {}

        obj = miniamf.MixedArray()
        self.context.addObject(obj)

        self.stream.read_ulong() #length

        self.readObjectAttributes(attrs)

        for key, value in attrs.items():
            try:
                key = int(key)
            except ValueError:
                pass

            obj[key] = value

        return obj

    cdef object readList(self):
        cdef list obj = []
        cdef unsigned long l
        cdef unsigned long i

        self.context.addObject(obj)
        l = self.stream.read_ulong()

        for i from 0 <= i < l:
            PyList_Append(obj, self.readElement())

        return obj

    cdef object readDate(self):
        cdef double ms = -1

        self.stream.read_double(&ms)
        self.stream.read_short() #timezone

        # Timezones are ignored
        d = util.get_datetime(ms / 1000.0)

        if self.timezone_offset:
            d = d + self.timezone_offset

        self.context.addObject(d)

        return d

    cdef object readLongString(self, bint bytes=0):
        cdef unsigned long l = 0
        cdef char *b = NULL
        cdef object s

        l = self.stream.read_ulong()

        self.stream.read(&b, l)
        s = PyBytes_FromStringAndSize(b, <Py_ssize_t>l)

        if bytes:
            return s

        return self.context.getStringForBytes(s)

    cdef object readXML(self):
        cdef object root = xml.fromstring(
            self.readLongString(),
            forbid_dtd=self.context.forbid_dtd,
            forbid_entities=self.context.forbid_entities
        )

        self.context.addObject(root)

        return root

    cdef object readAMF3(self):
        if self.amf3_decoder is None:
            self.context.amf3_context = amf3.Context()

            self.amf3_decoder = amf3.Decoder(
                stream=self.stream,
                context=self.context.amf3_context,
                timezone_offset=self.timezone_offset)

        return self.amf3_decoder.readElement()

    cdef object readConcreteElement(self, char type):
        if type == TYPE_NUMBER:
            return self.readNumber()
        elif type == TYPE_BOOL:
            return self.readBoolean()
        elif type == TYPE_STRING:
            return self.readString()
        elif type == TYPE_OBJECT:
            return self.readObject()
        elif type == TYPE_NULL:
            return self.readNull()
        elif type == TYPE_UNDEFINED:
            return self.readUndefined()
        elif type == TYPE_REFERENCE:
            return self.readReference()
        elif type == TYPE_MIXEDARRAY:
            return self.readMixedArray()
        elif type == TYPE_ARRAY:
            return self.readList()
        elif type == TYPE_DATE:
            return self.readDate()
        elif type == TYPE_LONGSTRING:
            return self.readLongString()
        elif type == TYPE_UNSUPPORTED:
            return self.readNull()
        elif type == TYPE_XML:
            return self.readXML()
        elif type == TYPE_TYPEDOBJECT:
            return self.readTypedObject()
        elif type == TYPE_AMF3:
            return self.readAMF3()

        raise miniamf.DecodeError("Unsupported ActionScript type")


cdef class Encoder(codec.Encoder):
    """
    The AMF0 Encoder.
    """

    def __cinit__(self):
        self.use_amf3 = 0

    def __init__(self, *args, **kwargs):
        self.use_amf3 = kwargs.pop('use_amf3', 0)

        self.context = kwargs.pop('context', None)

        if self.context is None:
            self.context = Context(**kwargs)

        codec.Codec.__init__(self, *args, **kwargs)

    cdef inline int writeReference(self, o) except -2:
        """
        Write reference to the data stream.
        """
        cdef Py_ssize_t idx = self.context.getObjectReference(o)

        if idx == -1 or idx > 65535:
            return -1

        self.writeType(TYPE_REFERENCE)

        return self.stream.write_ushort(idx)

    cdef int writeBoolean(self, b) except -1:
        self.writeType(TYPE_BOOL)

        if b is True:
            return self.writeType('\x01')

        return self.writeType('\x00')

    cdef int writeUndefined(self, data) except -1:
        return self.writeType(TYPE_UNDEFINED)

    cdef int writeNull(self, n) except -1:
        """
        Write null type to data stream.
        """
        return self.writeType(TYPE_NULL)

    cpdef int writeList(self, object a) except -1:
        """
        Write array to the stream.
        """
        cdef Py_ssize_t size = -1, i = -1

        if self.writeReference(a) != -1:
            return 0

        self.context.addObject(a)

        self.writeType(TYPE_ARRAY)
        size = PyList_GET_SIZE(a)

        self.stream.write_ulong(size)

        for i from 0 <= i < size:
            self.writeElement(<object>PyList_GET_ITEM(a, i))

        return 0

    cdef int writeTuple(self, object a) except -1:
        cdef Py_ssize_t size = -1, i = -1

        if self.writeReference(a) != -1:
            return 0

        self.context.addObject(a)

        self.writeType(TYPE_ARRAY)
        size = PyTuple_GET_SIZE(a)

        self.stream.write_ulong(size)

        for i from 0 <= i < size:
            self.writeElement(<object>PyTuple_GET_ITEM(a, i))

        return 0

    cdef int writeInt(self, object a) except -1:
        self.writeType(TYPE_NUMBER)

        return self.stream.write_double(a)

    cdef int writeNumber(self, n) except -1:
        self.writeType(TYPE_NUMBER)

        return self.stream.write_double(n)

    cdef int writeLong(self, object a) except -1:
        self.writeType(TYPE_NUMBER)

        return self.stream.write_double(a)

    cdef int writeBytes(self, s) except -1:
        """
        Write a string of bytes to the data stream.
        """
        cdef Py_ssize_t l = PyBytes_GET_SIZE(s)

        if l > 0xffff:
            self.writeType(TYPE_LONGSTRING)
        else:
            self.writeType(TYPE_STRING)

        if l > 0xffff:
            self.stream.write_ulong(l)
        else:
            self.stream.write_ushort(l)

        return self.stream.write(PyBytes_FromString(s), l)

    cdef int writeString(self, u) except -1:
        """
        Write a unicode to the data stream.
        """
        cdef object s = self.context.getBytesForString(u)

        return self.writeBytes(s)

    cpdef int serialiseString(self, u) except -1:
        """
        Similar to L{writeString} but does not encode a type byte.
        """
        if PyUnicode_CheckExact(u):
            u = self.context.getBytesForString(u)

        cdef Py_ssize_t l = PyBytes_GET_SIZE(u)

        if l > 0xffff:
            self.stream.write_ulong(l)
        else:
            self.stream.write_ushort(l)

        return self.stream.write(PyBytes_AsString(u), l)

    cdef int writeXML(self, e) except -1:
        """
        Writes an XML instance.
        """
        self.writeType(TYPE_XML)

        data = xml.tostring(e).decode('utf-8')

        if not PyUnicode_CheckExact(data):
            raise TypeError('expected str from xml.tostring')

        cdef Py_ssize_t l = PyUnicode_GetLength(data)

        self.stream.write_ulong(l)

        return self.stream.write(PyUnicode_AsUTF8String(data), l) #!Might be wrong

    cdef int writeDateTime(self, d) except -1:
        if self.timezone_offset is not None:
            d -= self.timezone_offset

        secs = util.get_timestamp(d)

        self.writeType(TYPE_DATE)
        self.stream.write_double(secs * 1000.0)

        return self.stream.write('\x00\x00', 2)

    cdef int writeDict(self, dict o) except -1:
        if self.writeReference(o) != -1:
            return 0

        self.context.addObject(o)
        self.writeType(TYPE_OBJECT)
        self._writeDict(o)

        return self._writeEndObject()

    cdef int _writeDict(self, dict attrs) except -1:
        """
        Write C{dict} to the data stream.

        @param o: The C{dict} data to be encoded to the AMF0 data stream.
        """
        for key, value in attrs.items():
            if PyLong_Check(key):
                key = str(key)

            self.serialiseString(key)
            self.writeElement(value)

        return 0

    cdef inline int _writeEndObject(self) except -1:
        return self.stream.write('\x00\x00\x09', 3)

    cpdef int writeObject(self, o) except -1:
        """
        Write a Python object to the stream.

        @param o: The object data to be encoded to the AMF0 data stream.
        """
        if self.writeReference(o) != -1:
            return 0

        self.context.addObject(o)
        alias = self.context.getClassAlias(o.__class__)

        alias.compile()

        if alias.amf3:
            return self.writeAMF3(o)

        if alias.anonymous:
            self.writeType(TYPE_OBJECT)
        else:
            self.writeType(TYPE_TYPEDOBJECT)
            self.serialiseString(alias.alias)

        cdef dict attrs = alias.getEncodableAttributes(o, codec=self)

        if len(attrs) and alias.static_attrs:
            for key in alias.static_attrs:
                value = attrs.pop(key)

                self.serialiseString(key)
                self.writeElement(value)

        if attrs:
            self._writeDict(attrs)

        return self._writeEndObject()

    cdef int writeMixedArray(self, o) except -1:
        if self.writeReference(o) != -1:
            return 0

        self.context.addObject(o)
        self.writeType(TYPE_MIXEDARRAY)

        # TODO: optimise this
        # work out the highest integer index
        try:
            # list comprehensions to save the day
            max_index = max([y for y in o.keys()
                if isinstance(y, int)])

            if max_index < 0:
                max_index = 0
        except ValueError:
            max_index = 0

        self.stream.write_ulong(max_index)

        self._writeDict(dict(o))
        self._writeEndObject()

    cdef int writeAMF3(self, o) except -1:
        if self.amf3_encoder is None:
            self.context.amf3_context = amf3.Context()

            self.amf3_encoder = amf3.Encoder(
                stream=self.stream,
                context=self.context.amf3_context,
                timezone_offset=self.timezone_offset)

        self.writeType(TYPE_AMF3)
        self.amf3_encoder.writeElement(o)

    cdef inline int handleBasicTypes(self, object element, object py_type) except -1:
        if self.use_amf3:
            return self.writeAMF3(element)

        return codec.Encoder.handleBasicTypes(self, element, py_type)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
The adapter package provides additional functionality for other Python
packages. This includes registering classes, setting up type maps etc.

@since: 0.1.0
"""

import os.path
import glob

from miniamf.util import imports
import importlib.resources as importlib_resources
from contextlib import ExitStack
import atexit


__all__ = [
    'register_adapters',
    'register_adapter',
    'get_adapter',
]


adapters_registered = False


class PackageImporter(object):
    """
    Package importer used for lazy module loading.
    """
    def __init__(self, name):
        self.name = name

    def __call__(self, mod):
        __import__(f'miniamf.adapters.{self.name}')


def register_adapters():
    """
    Register all adapters provided by submodules.  Submodules that
    fail to load are silently ignored, in case they depend on
    third-party packages that are not installed.
    """
    global adapters_registered

    if adapters_registered is True:
        return

    file_manager = ExitStack()
    atexit.register(file_manager.close)
    ref = importlib_resources.files('miniamf') / 'adapters'
    packageDir = file_manager.enter_context(importlib_resources.as_file(ref))

    for f in glob.glob(os.path.join(packageDir, '*.py')):
        mod = os.path.basename(f).split(os.path.extsep, 1)[0]

        if mod == '__init__' or not mod.startswith('_'):
            continue

        try:
            register_adapter(mod[1:].replace('_', '.'), PackageImporter(mod))
        except ImportError:
            pass

    adapters_registered = True


def register_adapter(mod, func):
    """
    Registers a callable to be executed when a module is imported. If the
    module already exists then the callable will be executed immediately.
    You can register the same module multiple times, the callables will be
    executed in the order they were registered. The root module must exist
    (i.e. be importable) otherwise an `ImportError` will be thrown.

    @param mod: The fully qualified module string, as used in the imports
        statement. E.g. 'foo.bar.baz'. The string must map to a module
        otherwise the callable will not fire.
    @param func: The function to call when C{mod} is imported. This function
        must take one arg, the newly imported C{module} object.
    @type func: callable
    @raise TypeError: C{func} must be callable
    """
    if not hasattr(func, '__call__'):
        raise TypeError('func must be callable')

    imports.when_imported(mod, func)


def get_adapter(mod):
    """
    Return the Mini-AMF adapter for the supplied module.

    Usage::
        adapter = miniamf.get_adapter('collections')
    """
    base_name = '_' + mod.replace('.', '_')

    full_import = f'{__name__}.{base_name}'

    ret = __import__(full_import)

    for attr in full_import.split('.')[1:]:
        ret = getattr(ret, attr)

    return ret



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests for the L{array} L{miniamf.adapters._array} module.

@since: 0.5
"""

import array
import unittest

import miniamf


class ArrayTestCase(unittest.TestCase):
    """
    """

    def setUp(self):
        self.orig = [ord('f'), ord('o'), ord('o')]

        self.obj = array.array('b')

        self.obj.append(ord('f'))
        self.obj.append(ord('o'))
        self.obj.append(ord('o'))

    def encdec(self, encoding):
        return next(miniamf.decode(
            miniamf.encode(self.obj, encoding=encoding),
            encoding=encoding))

    def test_amf0(self):
        self.assertEqual(self.encdec(miniamf.AMF0), self.orig)

    def test_amf3(self):
        self.assertEqual(self.encdec(miniamf.AMF3), self.orig)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests for AMF utilities.

@since: 0.1.0
"""

import unittest

import miniamf
from miniamf import codec


class TestObject(object):
    def __init__(self):
        self.name = 'test'


class DummyAlias(miniamf.ClassAlias):
    pass


class IndexedCollectionTestCase(unittest.TestCase):
    """
    Tests for L{codec.IndexedCollection}
    """

    def setUp(self):
        self.collection = codec.IndexedCollection()

    def test_clear(self):
        o = object()

        self.assertEqual(self.collection.getReferenceTo(o), -1)

        self.collection.append(o)
        self.assertEqual(self.collection.getReferenceTo(o), 0)

        self.collection.clear()

        self.assertEqual(self.collection.getReferenceTo(o), -1)

    def test_append(self):
        n = 5

        for i in range(0, n):
            test_obj = TestObject()

            test_obj.name = i

            self.collection.append(test_obj)

        self.assertEqual(len(self.collection), n)

        for i in range(0, n):
            self.assertEqual(i, self.collection[i].name)

    def test_get_reference_to(self):
        test_obj = TestObject()

        self.collection.append(test_obj)

        idx = self.collection.getReferenceTo(test_obj)

        self.assertEqual(0, idx)
        self.assertEqual(-1, self.collection.getReferenceTo(TestObject()))

    def test_get_by_reference(self):
        test_obj = TestObject()
        idx = self.collection.append(test_obj)

        self.assertIdentical(test_obj, self.collection.getByReference(idx))

        idx = self.collection.getReferenceTo(test_obj)

        self.assertIdentical(test_obj, self.collection.getByReference(idx))
        self.assertRaises(TypeError, self.collection.getByReference, 'bad ref')

        self.assertEqual(None, self.collection.getByReference(74))

    def test_len(self):
        self.assertEqual(0, len(self.collection))

        self.collection.append([])

        self.assertEqual(1, len(self.collection))

        self.collection.append({})

        self.assertEqual(2, len(self.collection))

        self.collection.clear()
        self.assertEqual(0, len(self.collection))

    def test_repr(self):
        x = "0x%x" % id(self.collection)

        self.assertEqual(
            repr(self.collection),
            '<miniamf.codec.IndexedCollection size=0 %s>' % (x,)
        )

    def test_contains(self):
        o = object()

        self.assertFalse(o in self.collection)

        self.collection.append(o)

        self.assertTrue(o in self.collection)

    def test_eq(self):
        self.assertEqual(self.collection, [])
        self.assertRaises(NotImplementedError, self.collection.__eq__, self)

    def test_hash(self):
        class A(object):
            def __hash__(self):
                return 1

        self.collection = codec.IndexedCollection(True)

        o = A()

        self.assertEqual(self.collection.getReferenceTo(o), -1)

        self.collection.append(o)
        self.assertEqual(self.collection.getReferenceTo(o), 0)

        self.collection.clear()

        self.assertEqual(self.collection.getReferenceTo(o), -1)


class ContextTestCase(unittest.TestCase):
    """
    Tests for L{codec.Context}
    """

    def setUp(self):
        self.context = codec.Context()

    def test_add(self):
        y = [1, 2, 3]

        self.assertEqual(self.context.getObjectReference(y), -1)
        self.assertEqual(self.context.addObject(y), 0)
        self.assertEqual(self.context.getObjectReference(y), 0)

    def test_clear(self):
        y = [1, 2, 3]

        self.context.addObject(y)

        self.assertEqual(self.context.getObjectReference(y), 0)

        self.context.clear()

        self.assertEqual(self.context.getObjectReference(y), -1)

    def test_get_by_reference(self):
        y = [1, 2, 3]
        z = {'spam': 'eggs'}

        self.context.addObject(y)
        self.context.addObject(z)

        self.assertIdentical(self.context.getObject(0), y)
        self.assertIdentical(self.context.getObject(1), z)
        self.assertIdentical(self.context.getObject(2), None)

        for t in ['', 2.2323]:
            self.assertRaises(TypeError, self.context.getObject, t)

    def test_get_reference(self):
        y = [1, 2, 3]
        z = {'spam': 'eggs'}

        ref1 = self.context.addObject(y)
        ref2 = self.context.addObject(z)

        self.assertIdentical(self.context.getObjectReference(y), ref1)
        self.assertIdentical(self.context.getObjectReference(z), ref2)
        self.assertEqual(self.context.getObjectReference({}), -1)

    def test_no_alias(self):
        class A:
            pass

        alias = self.context.getClassAlias(A)

        self.assertTrue(isinstance(alias, miniamf.ClassAlias))
        self.assertIdentical(alias.klass, A)

    def test_registered_alias(self):
        class A:
            pass

        miniamf.register_class(A)
        self.addCleanup(miniamf.unregister_class, A)

        alias = self.context.getClassAlias(A)

        self.assertTrue(isinstance(alias, miniamf.ClassAlias))
        self.assertIdentical(alias.klass, A)

    def test_registered_deep(self):
        class A:
            pass

        class B(A):
            pass

        miniamf.register_alias_type(DummyAlias, A)
        self.addCleanup(miniamf.unregister_alias_type, DummyAlias)
        alias = self.context.getClassAlias(B)

        self.assertTrue(isinstance(alias, DummyAlias))
        self.assertIdentical(alias.klass, B)

    def test_get_class_alias(self):
        class A:
            pass

        alias1 = self.context.getClassAlias(A)
        alias2 = self.context.getClassAlias(A)

        self.assertIdentical(alias1, alias2)

    def test_string(self):
        s = b'foo'
        u = self.context.getStringForBytes(s)

        self.assertTrue(isinstance(u, str))
        self.assertEqual(u, s.decode('ascii'))

        i = self.context.getStringForBytes(s)

        self.assertIdentical(u, i)

        self.context.clear()

        i = self.context.getStringForBytes(s)

        self.assertFalse(u is i)

    def test_bytes(self):
        s = u'foo'

        b = self.context.getBytesForString(s)

        self.assertTrue(isinstance(b, bytes))
        self.assertEqual(b, s.encode('ascii'))

        i = self.context.getBytesForString(s)

        self.assertIdentical(i, b)

        self.context.clear()

        i = self.context.getBytesForString(s)

        self.assertNotIdentical(i, s)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
C-extension for L{miniamf.amf3} Python module in L{Mini-AMF<miniamf>}.

@since: 0.6
"""

from cpython cimport *
from libc.stdlib cimport malloc, free

cimport cython

from miniamf._accel.util cimport cBufferedByteStream, BufferedByteStream
from miniamf._accel cimport codec
import miniamf
from miniamf import util, amf3, xml
import types


try:
    import zlib
except ImportError:
    zlib = None


cdef char TYPE_UNDEFINED = '\x00'
cdef char TYPE_NULL = '\x01'
cdef char TYPE_BOOL_FALSE = '\x02'
cdef char TYPE_BOOL_TRUE = '\x03'
cdef char TYPE_INTEGER = '\x04'
cdef char TYPE_NUMBER = '\x05'
cdef char TYPE_STRING = '\x06'
cdef char TYPE_XML = '\x07'
cdef char TYPE_DATE = '\x08'
cdef char TYPE_ARRAY = '\x09'
cdef char TYPE_OBJECT = '\x0A'
cdef char TYPE_XMLSTRING = '\x0B'
cdef char TYPE_BYTEARRAY = '\x0C'

cdef unsigned int REFERENCE_BIT = 0x01
cdef char REF_CHAR = '\x01'

#: The maximum that can be represented by an signed 29 bit integer.
cdef long MAX_29B_INT = 0x0FFFFFFF

#: The minimum that can be represented by an signed 29 bit integer.
cdef long MIN_29B_INT = -0x10000000

cdef int OBJECT_ENCODING_STATIC = 0x00
cdef int OBJECT_ENCODING_EXTERNAL = 0x01
cdef int OBJECT_ENCODING_DYNAMIC = 0x02

cdef object ByteArrayType = amf3.ByteArray
cdef object DataInput = amf3.DataInput
cdef object DataOutput = amf3.DataOutput
cdef str empty_string = str('')
cdef unicode empty_unicode = ''
cdef object undefined = miniamf.Undefined


cdef class ClassDefinition(object):
    """
    Holds transient class trait info for an individual encode/decode.
    """

    def __cinit__(self):
        self.alias = None
        self.ref = -1
        self.attr_len = -1
        self.encoding = -1
        self.encoded_ref = NULL
        self.encoded_ref_size = -1

    def __init__(self, alias):
        self.alias = alias

        alias.compile()

        self.attr_len = 0
        self.static_properties = []

        if alias.static_attrs:
            self.attr_len = len(alias.static_attrs)
            self.static_properties = alias.static_attrs

        self.encoding = OBJECT_ENCODING_DYNAMIC

        if alias.external:
            self.encoding = OBJECT_ENCODING_EXTERNAL
        elif not alias.dynamic:
            if alias.encodable_properties is not None:
                if len(alias.static_attrs) == len(alias.encodable_properties):
                    self.encoding = OBJECT_ENCODING_STATIC
            else:
                self.encoding = OBJECT_ENCODING_STATIC

    def __dealloc__(self):
        if self.encoded_ref != NULL:
            free(self.encoded_ref)
            self.encoded_ref = NULL

    cdef int writeReference(self, cBufferedByteStream stream):
        if self.encoded_ref != NULL:
            return stream.write(self.encoded_ref, self.encoded_ref_size)

        cdef Py_ssize_t ref = 0
        cdef char *buf = NULL

        if self.encoding != OBJECT_ENCODING_EXTERNAL:
            ref += self.attr_len << 4

        ref |= self.encoding << 2 | REFERENCE_BIT << 1 | REFERENCE_BIT

        try:
            stream.write(buf, encode_int(ref, &buf))
        finally:
            if buf != NULL:
                free(buf)

        try:
            self.encoded_ref_size = encode_int(self.ref << 2 | REFERENCE_BIT, &self.encoded_ref)
        except:
            if self.encoded_ref != NULL:
                free(self.encoded_ref)
                self.encoded_ref = NULL

            raise

        return 0


cdef class Context(codec.Context):
    """
    I hold the AMF3 context for en/decoding streams.
    """

    def __cinit__(self):
        self.strings = codec.ByteStringReferenceCollection()
        self.classes = {}
        self.class_ref = {}

        self.class_idx = 0

    cpdef int clear(self) except -1:
        """
        Clears the context.
        """
        codec.Context.clear(self)

        self.strings.clear()

        self.classes = {}
        self.class_ref = {}
        self.class_idx = 0

        return 0

    cpdef object getString(self, Py_ssize_t ref):
        return self.strings.getByReference(ref)

    cpdef Py_ssize_t getStringReference(self, object s) except -2:
        return self.strings.getReferenceTo(s)

    cpdef Py_ssize_t addString(self, object s) except -1:
        """
        Returns -2 which signifies that s was empty
        """
        return self.strings.append(s)

    cpdef object getClassByReference(self, Py_ssize_t ref):
        return self.class_ref.get(ref, None)

    cpdef ClassDefinition getClass(self, object klass):
        return self.classes.get(klass, None)

    cpdef Py_ssize_t addClass(self, ClassDefinition alias, klass) except? -1:
        cdef object ref = self.class_idx

        self.class_ref[ref] = alias
        self.classes[klass] = alias

        alias.ref = ref
        self.class_idx += 1

        return ref


cdef class Decoder(codec.Decoder):
    """
    Decodes an AMF3 data stream.
    """

    def __init__(self, *args, **kwargs):
        context = kwargs.pop('context', None)

        if context is None:
            context = Context()

        self.context = context

        codec.Decoder.__init__(self, *args, **kwargs)

    cdef object readInteger(self, int signed=1):
        """
        Reads and returns an integer from the stream.

        @type signed: C{bool}
        @see: U{Parsing integers on OSFlash
        <http://osflash.org/amf3/parsing_integers>} for the AMF3 integer data
        format.
        """
        return <object>decode_int(self.stream, signed)

    cdef object readNumber(self):
        cdef double d = -1

        self.stream.read_double(&d)

        return d

    cdef object readBytes(self):
        cdef object s = self.readString()

        return self.context.getBytesForString(s)

    cpdef object readString(self):
        """
        Reads and returns a string from the stream.
        """
        cdef Py_ssize_t r = _read_ref(self.stream)
        cdef object s

        if r & REFERENCE_BIT == 0:
            # read a string reference
            return self.context.getString(r >> 1)

        r >>= 1

        if r == 0:
            return empty_unicode

        cdef char *buf = NULL

        self.stream.read(&buf, r)
        s = PyUnicode_DecodeUTF8(buf, r, 'strict')

        self.context.addString(s)

        return s

    cdef object readDate(self):
        """
        Read date from the stream.

        The timezone is ignored as the date is always in UTC.
        """
        cdef Py_ssize_t ref = _read_ref(self.stream)

        if ref & REFERENCE_BIT == 0:
            return self.context.getObject(ref >> 1)

        cdef double ms

        self.stream.read_double(&ms)

        cdef object result = util.get_datetime(ms / 1000.0)

        if self.timezone_offset is not None:
            result += self.timezone_offset

        self.context.addObject(result)

        return result

    cdef object readList(self):
        """
        Reads an array from the stream.
        """
        cdef int size = _read_ref(self.stream)
        cdef int i
        cdef list result
        cdef object tmp
        cdef unicode key

        if size & REFERENCE_BIT == 0:
            return self.context.getObject(size >> 1)

        size >>= 1
        key = self.readString()

        if PyUnicode_GetLength(key) == 0:
            # integer indexes only -> python list
            result = []
            self.context.addObject(result)

            for i from 0 <= i < size:
                result.append(self.readElement())

            return result

        tmp = miniamf.MixedArray()
        self.context.addObject(tmp)

        while PyUnicode_GetLength(key):
            tmp[key] = self.readElement()
            key = self.readString()

        for i from 0 <= i < size:
            el = self.readElement()
            tmp[i] = el

        return tmp

    cdef ClassDefinition _getClassDefinition(self, long ref):
        """
        Reads class definition from the stream.
        """
        if ref & REFERENCE_BIT == 0:
            return self.context.getClassByReference(ref >> 1)

        ref >>= 1

        cdef object name = self.readString()
        cdef object alias = None
        cdef Py_ssize_t i

        if PyUnicode_GetLength(name) == 0:
            name = miniamf.ASObject

        try:
            alias = self.context.getClassAlias(name)
        except miniamf.UnknownClassAlias:
            if self.strict:
                raise

            alias = miniamf.TypedObjectClassAlias(name)

        cdef ClassDefinition class_def = ClassDefinition(alias)

        class_def.encoding = ref & 0x03
        class_def.attr_len = ref >> 2
        class_def.static_properties = []

        if class_def.attr_len > 0:
            for i from 0 <= i < class_def.attr_len:
                class_def.static_properties.append(self.readString())

        self.context.addClass(class_def, alias.klass)

        return class_def

    @cython.boundscheck(False)
    cdef int _readStatic(self, ClassDefinition class_def, dict obj) except -1:
        cdef Py_ssize_t i

        for 0 <= i < class_def.attr_len:
            obj[class_def.static_properties[i]] = self.readElement()

        return 0

    cdef int _readDynamic(self, ClassDefinition class_def, dict obj) except -1:
        cdef object attr
        cdef char *peek = NULL


        while True:
            self.stream.peek(&peek, 1)

            if peek[0] == REF_CHAR:
                self.stream.seek(1, 1)

                break

            PyDict_SetItem(obj, self.readString(), self.readElement())

        return 0

    cdef object readObject(self):
        """
        Reads an object from the stream.

        @raise miniamf.EncodeError: Decoding an object in amf3 tagged as amf0
            only is not allowed.
        @raise miniamf.DecodeError: Unknown object encoding.
        """
        cdef int ref = _read_ref(self.stream)
        cdef object obj

        if ref & REFERENCE_BIT == 0:
            obj = self.context.getObject(ref >> 1)

            if obj is None:
                raise miniamf.ReferenceError('Unknown reference')

            return obj

        cdef ClassDefinition class_def = self._getClassDefinition(ref >> 1)
        cdef object alias = class_def.alias

        obj = alias.createInstance(codec=self)
        cdef dict obj_attrs = {}

        self.context.addObject(obj)

        if class_def.encoding == OBJECT_ENCODING_DYNAMIC:
            self._readStatic(class_def, obj_attrs)
            self._readDynamic(class_def, obj_attrs)
        elif class_def.encoding == OBJECT_ENCODING_STATIC:
            self._readStatic(class_def, obj_attrs)
        elif class_def.encoding == OBJECT_ENCODING_EXTERNAL:
            obj.__readamf__(DataInput(self))

            return obj
        else:
            raise miniamf.DecodeError("Unknown object encoding")

        alias.applyAttributes(obj, obj_attrs, codec=self)

        return obj

    cdef object readXML(self):
        """
        Reads an XML object from the stream.
        """
        cdef int ref = _read_ref(self.stream)

        if ref & REFERENCE_BIT == 0:
            return self.context.getObject(ref >> 1)

        ref >>= 1

        cdef char *buf = NULL

        self.stream.read(&buf, ref)

        x = xml.fromstring(
            PyBytes_FromStringAndSize(buf, ref),
            forbid_dtd=self.context.forbid_dtd,
            forbid_entities=self.context.forbid_entities
        )
        self.context.addObject(x)

        return x

    cdef object readByteArray(self):
        """
        Reads a string of data from the stream.

        Detects if the L{ByteArray} was compressed using C{zlib}.

        @see: L{ByteArray}
        @note: This is not supported in ActionScript 1.0 and 2.0.
        """
        cdef int ref = _read_ref(self.stream)

        if ref & REFERENCE_BIT == 0:
            return self.context.getObject(ref >> 1)

        cdef char *buf = NULL
        cdef object s
        cdef object compressed = None

        ref >>= 1

        self.stream.read(&buf, ref)
        s = PyBytes_FromStringAndSize(buf, ref)

        if zlib:
            if ref > 2 and buf[0] == '\x78' and buf[1] == '\x9c':
                try:
                    s = zlib.decompress(s)
                except zlib.error:
                    pass

        s = (<object>ByteArrayType)(s)

        self.context.addObject(s)

        return s

    cdef object readConcreteElement(self, char t):
        if t == TYPE_STRING:
            return self.readString()
        elif t == TYPE_OBJECT:
            return self.readObject()
        elif t == TYPE_UNDEFINED:
            return undefined
        elif t == TYPE_NULL:
            return None
        elif t == TYPE_BOOL_FALSE:
            return False
        elif t == TYPE_BOOL_TRUE:
            return True
        elif t == TYPE_INTEGER:
            return self.readInteger(1)
        elif t == TYPE_NUMBER:
            return self.readNumber()
        elif t == TYPE_ARRAY:
            return self.readList()
        elif t == TYPE_DATE:
            return self.readDate()
        elif t == TYPE_BYTEARRAY:
            return self.readByteArray()
        elif t == TYPE_XML:
            return self.readXML()
        elif t == TYPE_XMLSTRING:
            return self.readXML()

        raise miniamf.DecodeError("Unsupported ActionScript type")


cdef class Encoder(codec.Encoder):
    """
    The AMF3 Encoder.
    """

    def __init__(self, *args, **kwargs):
        context = kwargs.pop('context', None)

        if context is None:
            context = Context()

        self.context = context

        codec.Encoder.__init__(self, *args, **kwargs)

    cdef inline int writeNull(self, n) except -1:
        """
        Writes a C{null} value to the stream.
        """
        return self.writeType(TYPE_NULL)

    cdef inline int writeUndefined(self, n) except -1:
        return self.writeType(TYPE_UNDEFINED)

    cdef inline int writeBoolean(self, n) except -1:
        if n is True:
            return self.writeType(TYPE_BOOL_TRUE)

        return self.writeType(TYPE_BOOL_FALSE)

    cpdef int serialiseString(self, u) except -1:
        """
        Serialises a unicode object.
        """
        cdef Py_ssize_t l
        cdef bint is_unicode = 0

        if PyUnicode_Check(u):
            l = PyUnicode_GetLength(u)
            is_unicode = 1
        elif PyBytes_Check(u):
            l = PyBytes_GET_SIZE(u)
        else:
            raise TypeError('Expected str or unicode')

        if l == 0:
            # '' is a special case
            return self.stream.write(&REF_CHAR, 1)

        r = self.context.getStringReference(u)

        if r != -1:
            # we have a reference
            return _encode_integer(self.stream, r << 1)

        self.context.addString(u)

        if is_unicode:
            u = self.context.getBytesForString(u)
            l = PyBytes_GET_SIZE(u)

        _encode_integer(self.stream, (l << 1) | REFERENCE_BIT)

        return self.stream.write(PyBytes_AsString(u), l)

    cdef int writeString(self, object s) except -1:
        self.writeType(TYPE_STRING)
        self.serialiseString(s)

    cdef int writeBytes(self, object s) except -1:
        self.writeType(TYPE_STRING)
        self.serialiseString(s)

    cdef int writeInt(self, object n) except -1:
        cdef long x = PyLong_AsLong(n)

        if x < MIN_29B_INT or x > MAX_29B_INT:
            return self.writeNumber(float(n))

        self.writeType(TYPE_INTEGER)
        _encode_integer(self.stream, x)

    cdef int writeLong(self, object n) except -1:
        cdef long x

        try:
            x = PyLong_AsLong(n)
        except:
            return self.writeNumber(float(n))

        if x < MIN_29B_INT or x > MAX_29B_INT:
            return self.writeNumber(float(n))

        self.writeType(TYPE_INTEGER)
        _encode_integer(self.stream, x)

    cdef int writeNumber(self, object n) except -1:
        cdef double x = PyFloat_AS_DOUBLE(n)

        self.writeType(TYPE_NUMBER)
        self.stream.write_double(x)

        return 0

    cpdef int writeList(self, object n) except -1:
        cdef Py_ssize_t ref = self.context.getObjectReference(n)
        cdef Py_ssize_t i

        self.writeType(TYPE_ARRAY)

        if ref != -1:
            return _encode_integer(self.stream, ref << 1)

        self.context.addObject(n)
        ref = PyList_GET_SIZE(n)

        _encode_integer(self.stream, (ref << 1) | REFERENCE_BIT)

        self.writeType('\x01')

        for i from 0 <= i < ref:
            self.writeElement(<object>PyList_GET_ITEM(n, i))

        return 0

    cdef int writeTuple(self, object n) except -1:
        cdef Py_ssize_t ref = self.context.getObjectReference(n)
        cdef Py_ssize_t i

        self.writeType(TYPE_ARRAY)

        if ref != -1:
            return _encode_integer(self.stream, ref << 1)

        self.context.addObject(n)

        ref = PyTuple_GET_SIZE(n)

        _encode_integer(self.stream, (ref << 1) | REFERENCE_BIT)
        self.writeType('\x01')

        for i from 0 <= i < ref:
            self.writeElement(<object>PyTuple_GET_ITEM(n, i))

        return 0

    cdef int writeDict(self, dict obj) except -1:
        if '' in obj:
            raise miniamf.EncodeError("dicts cannot contain empty string keys")

        # Pure python writes TYPE_ARRAY here but that doesn't work here
        self.writeType(TYPE_OBJECT)

        ref = self.context.getObjectReference(obj)

        if ref != -1:
            _encode_integer(self.stream, ref << 1)

            return 0

        self.context.addObject(obj)

        cdef bint class_ref = 0
        cdef ClassDefinition definition = self.context.getClass(dict)

        if not definition:
            definition = ClassDefinition(self.context.getClassAlias(dict))
            self.context.addClass(definition, dict)
        else:
            class_ref = 1

        definition.writeReference(self.stream)

        int_keys = PyList_New(0)
        str_keys = PyList_New(0)

        for x in obj.keys():
            if isinstance(x, int):
                PyList_Append(int_keys, x)
            elif isinstance(x, str):
                PyList_Append(str_keys, x)
            else:
                raise ValueError("Non integer/string key value found in dict")

        # Make sure the integer keys are within range
        l = PyList_Size(int_keys)

        for x in int_keys:
            if l < x <= 0:
                # treat as a string key
                str_keys.append(x)
                del int_keys[int_keys.index(x)]

        PyList_Sort(int_keys)

        # If integer keys don't start at 0, they will be treated as strings
        if len(int_keys) > 0 and int_keys[0] != 0:
            for x in int_keys.copy():
                PyList_Append(str_keys, str(x))
                del int_keys[int_keys.index(x)]

        _encode_integer(self.stream, PyList_Size(int_keys) << 1 | REFERENCE_BIT)

        PyList_Sort(str_keys)

        for x in str_keys:
            self.serialiseString(x)
            try:
                self.writeElement(obj[x])
            except KeyError:
                self.writeElement(obj[int(x)])

        if class_ref == 0:
            self.stream.write(&REF_CHAR, 1)

        for k in int_keys:
            self.writeElement(obj[k])
    cdef int writeMixedArray(self, object n) except -1:
        # Design bug in AMF3 that cannot read/write empty key strings
        # http://www.docuverse.com/blog/donpark/2007/05/14/flash-9-amf3-bug
        # for more info
        if '' in n:
            raise miniamf.EncodeError("dicts cannot contain empty string keys")

        self.writeType(TYPE_ARRAY)

        ref = self.context.getObjectReference(n)

        if ref != -1:
            return _encode_integer(self.stream, ref << 1)

        self.context.addObject(n)

        # The AMF3 spec demands that all str based indicies be listed first
        int_keys = []
        str_keys = []

        for x in n.keys():
            if isinstance(x, int):
                int_keys.append(x)
            elif isinstance(x, (str, unicode)):
                str_keys.append(x)
            else:
                raise ValueError("Non int/str key value found in dict")

        # Make sure the integer keys are within range
        l = len(int_keys)

        for x in int_keys:
            if l < x <= 0:
                # treat as a string key
                str_keys.append(x)
                del int_keys[int_keys.index(x)]

        int_keys.sort()

        # If integer keys don't start at 0, they will be treated as strings
        if len(int_keys) > 0 and int_keys[0] != 0:
            for x in int_keys:
                str_keys.append(str(x))
                del int_keys[int_keys.index(x)]

        _encode_integer(self.stream, len(int_keys) << 1 | REFERENCE_BIT)

        for x in str_keys:
            self.serialiseString(x)
            self.writeElement(n[x])

        self.stream.write_uchar(0x01)

        for k in int_keys:
            self.writeElement(n[k])

    cpdef int writeObject(self, object obj) except -1:
        cdef Py_ssize_t ref
        cdef object kls
        cdef ClassDefinition definition
        cdef object alias = None
        cdef int class_ref = 0
        cdef char *buf = NULL
        cdef PyObject *key
        cdef PyObject *value
        cdef object attrs

        self.writeType(TYPE_OBJECT)

        ref = self.context.getObjectReference(obj)

        if ref != -1:
            _encode_integer(self.stream, ref << 1)

            return 0

        self.context.addObject(obj)

        # object is not referenced, serialise it
        kls = obj.__class__
        definition = self.context.getClass(kls)

        if definition:
            class_ref = 1
            alias = definition.alias
        else:
            alias = self.context.getClassAlias(kls)
            definition = ClassDefinition(alias)

            self.context.addClass(definition, alias.klass)

        definition.writeReference(self.stream)

        if class_ref == 0:
            if alias.anonymous:
                self.stream.write(&REF_CHAR, 1)
            else:
                self.serialiseString(alias.alias)

            # work out what the final reference for the class will be.
            # this is okay because the next time an object of the same
            # class is encoded, class_ref will be True and never get here
            # again.

        if alias.external:
            obj.__writeamf__(DataOutput(self))

            return 0

        attrs = alias.getEncodableAttributes(obj, codec=self)

        if PyDict_CheckExact(attrs) != 1:
            raise TypeError('Expected dict for encodable attributes')

        if definition.attr_len > 0:
            if class_ref == 0:
                for attr in definition.static_properties:
                    self.serialiseString(attr)

            for attr in definition.static_properties:
                value = PyDict_GetItem(attrs, attr)

                if value == NULL:
                    raise KeyError

                if PyDict_DelItem(attrs, attr) == -1:
                    return -1

                self.writeElement(<object>value)

            if definition.encoding == OBJECT_ENCODING_STATIC:
                return 0

        if definition.encoding == OBJECT_ENCODING_DYNAMIC:
            ref = 0
            key = NULL
            value = NULL

            while PyDict_Next(attrs, &ref, &key, &value):
                self.serialiseString(<object>key)
                self.writeElement(<object>value)

            self.stream.write(&REF_CHAR, 1)

        return 0

    cdef int writeByteArray(self, object obj) except -1:
        """
        Writes a L{ByteArray} to the data stream.

        @param n: The L{ByteArray} data to be encoded to the AMF3 data stream.
        @type n: L{ByteArray}
        """
        cdef Py_ssize_t ref
        cdef object buf

        self.writeType(TYPE_BYTEARRAY)

        ref = self.context.getObjectReference(obj)

        if ref != -1:
            _encode_integer(self.stream, ref << 1)

            return 0

        self.context.addObject(obj)

        buf = obj.encode()
        l = PyBytes_GET_SIZE(buf)

        _encode_integer(self.stream, (l << 1) | REFERENCE_BIT)
        self.stream.write(PyBytes_AsString(buf), l)

        return 0

    cdef int writeXML(self, obj) except -1:
        self.writeType(TYPE_XMLSTRING)

        i = self.context.getObjectReference(obj)

        if i != -1:
            _encode_integer(self.stream, i << 1)

            return 0

        self.context.addObject(obj)

        s = xml.tostring(obj).decode('utf-8')

        if not PyUnicode_CheckExact(s):
            raise TypeError('Expected string from xml serialization')

        i = PyUnicode_GetLength(s)

        _encode_integer(self.stream, (i << 1) | REFERENCE_BIT)
        self.stream.write(PyUnicode_AsUTF8String(s), i)

        return 0

    cdef int writeDateTime(self, obj) except -1:
        """
        Writes an L{datetime.datetime} object to the stream
        """
        cdef Py_ssize_t ref = self.context.getObjectReference(obj)

        self.writeType(TYPE_DATE)

        if ref != -1:
            _encode_integer(self.stream, ref << 1)

            return 0

        self.context.addObject(obj)
        self.stream.write(&REF_CHAR, 1)

        if self.timezone_offset is not None:
            obj -= self.timezone_offset

        ms = <double>util.get_timestamp(obj)
        self.stream.write_double(ms * 1000.0)

    cdef int handleBasicTypes(self, object element, object py_type) except -1:
        cdef int ret = codec.Encoder.handleBasicTypes(self, element, py_type)

        if ret == 1: # not handled
            if py_type is ByteArrayType:
                return self.writeByteArray(element)

        return ret


cdef int encode_int(long i, char **buf) except -1:
    # Use typecasting to get the twos complement representation of i
    cdef unsigned long n = (<unsigned long*>(<void *>(&i)))[0]

    cdef unsigned long real_value = n
    cdef unsigned char count = 0
    cdef char *bytes = NULL

    if n > 0x1fffff:
        bytes = <char *>malloc(4)

        if bytes == NULL:
            PyErr_NoMemory()

        n = n >> 1
        bytes[count] = 0x80 | ((n >> 21) & 0xff)
        count += 1

    if n > 0x3fff:
        if bytes == NULL:
            bytes = <char *>malloc(3)

            if bytes == NULL:
                PyErr_NoMemory()

        bytes[count] = 0x80 | ((n >> 14) & 0xff)
        count += 1

    if n > 0x7f:
        if bytes == NULL:
            bytes = <char *>malloc(2)

            if bytes == NULL:
                PyErr_NoMemory()

        bytes[count] = 0x80 | ((n >> 7) & 0xff)
        count += 1

    if bytes == NULL:
        bytes = <char *>malloc(1)

        if bytes == NULL:
            PyErr_NoMemory()

    if real_value > 0x1fffff:
        bytes[count] = real_value & 0xff
    else:
        bytes[count] = real_value & 0x7f

    buf[0] = bytes

    return count + 1


cdef int decode_int(cBufferedByteStream stream, int sign=0) except? -1:
    cdef int n = 0
    cdef int result = 0
    cdef unsigned char b = stream.read_uchar()

    while b & 0x80 != 0 and n < 3:
        result <<= 7
        result |= b & 0x7f

        b = stream.read_uchar()

        n += 1

    if n < 3:
        result <<= 7
        result |= b
    else:
        result <<= 8
        result |= b

        if result & 0x10000000 != 0:
            if sign == 1:
                result -= 0x20000000
            else:
                result <<= 1
                result += 1

    return result


cdef inline int _encode_integer(cBufferedByteStream stream, int i) except -1:
    cdef char *buf = NULL

    try:
        return stream.write(buf, encode_int(i, &buf))
    finally:
        free(buf)


cdef inline Py_ssize_t _read_ref(cBufferedByteStream stream) except -1:
    return <Py_ssize_t>decode_int(stream, 0)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests for the C{sets} module integration.
"""

import unittest

import miniamf


# All set types are mapped to sorted tuples.
class BaseTestCase(unittest.TestCase):
    def amf0_encode_test(self, ty):

        x = ty(['1', '2', '3'])
        self.assertEqual(
            miniamf.encode(x, encoding=miniamf.AMF0).getvalue(),
            b'\n\x00\x00\x00\x03'
            b'\x02\x00\x011'
            b'\x02\x00\x012'
            b'\x02\x00\x013'
        )

        y = ty(['z', 'x', 'c', 'v'])
        self.assertEqual(
            miniamf.encode(y, encoding=miniamf.AMF0).getvalue(),
            b'\n\x00\x00\x00\x04'
            b'\x02\x00\x01c'
            b'\x02\x00\x01v'
            b'\x02\x00\x01x'
            b'\x02\x00\x01z'
        )

    def amf3_encode_test(self, ty):

        x = ty(['1', '2', '3'])
        self.assertEqual(
            miniamf.encode(x, encoding=miniamf.AMF3).getvalue(),
            b'\t\x07\x01'
            b'\x06\x031'
            b'\x06\x032'
            b'\x06\x033'
        )

        y = ty(['z', 'x', 'c', 'v'])
        self.assertEqual(
            miniamf.encode(y, encoding=miniamf.AMF3).getvalue(),
            b'\t\t\x01'
            b'\x06\x03c'
            b'\x06\x03v'
            b'\x06\x03x'
            b'\x06\x03z'
        )


class BuiltinSetTypesTestCase(BaseTestCase):
    def test_amf0_set(self):
        self.amf0_encode_test(set)

    def test_amf3_set(self):
        self.amf3_encode_test(set)

    def test_amf0_frozenset(self):
        self.amf0_encode_test(frozenset)

    def test_amf3_frozenset(self):
        self.amf3_encode_test(frozenset)


# The sets module was removed in Python 3.
try:
    ModuleNotFoundError
except NameError:
    ModuleNotFoundError = ImportError

try:
    import sets

    class LibrarySetTypesTestCase(BaseTestCase):
        def test_amf0_Set(self):
            self.amf0_encode_test(sets.Set)

        def test_amf3_Set(self):
            self.amf3_encode_test(sets.Set)

        def test_amf0_ImmutableSet(self):
            self.amf0_encode_test(sets.ImmutableSet)

        def test_amf3_ImmutableSet(self):
            self.amf3_encode_test(sets.ImmutableSet)

except ModuleNotFoundError:
    pass



# cython: boundscheck=False
# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
C-extension for L{miniamf.amf3} Python module in L{Mini-AMF<miniamf>}.

@since: 0.4
"""

from cpython cimport *

cdef extern from "datetime.h":
    void PyDateTime_IMPORT()
    int PyDateTime_CheckExact(object)
    int PyDate_CheckExact(object)
    int PyTime_CheckExact(object)

from miniamf._accel.util cimport cBufferedByteStream, BufferedByteStream

import types
import miniamf
from miniamf import util, xml
import datetime


cdef object MixedArray = miniamf.MixedArray
cdef object Undefined = miniamf.Undefined
cdef object BuiltinFunctionType = types.BuiltinFunctionType
cdef object GeneratorType = types.GeneratorType

PyDateTime_IMPORT


cdef class IndexedCollection(object):
    """
    Provides reference functionality for amf contexts.

    @see: L{miniamf.codec.IndexedCollection} for complete documentation
    """

    def __cinit__(self, bint use_hash=0):
        self.use_hash = use_hash

        self.data = NULL
        self.refs = {}
        self.size = -1
        self.length = -1

    def __init__(self, use_hash=False):
        self.use_hash = use_hash

        self.clear()

    cdef void _clear(self):
        cdef Py_ssize_t i

        if self.data != NULL:
            for i from 0 <= i < self.length:
                Py_DECREF(<object>self.data[i])

            PyMem_Free(self.data)
            self.data = NULL

    def __dealloc__(self):
        self._clear()

    cdef int _actually_increase_size(self) except -1:
        cdef Py_ssize_t new_len = self.length
        cdef Py_ssize_t current_size = self.size
        cdef PyObject **cpy

        while new_len >= current_size:
            current_size *= 2

        if current_size != self.size:
            self.size = current_size

            cpy = <PyObject **>PyMem_Realloc(
                self.data,
                sizeof(PyObject *) * self.size
            )

            if cpy == NULL:
                self._clear()

                PyErr_NoMemory()

            self.data = cpy

        return 0

    cdef inline int _increase_size(self) except -1:
        if self.length < self.size:
            return 0

        return self._actually_increase_size()

    cpdef int clear(self) except -1:
        self._clear()

        self.length = 0
        self.size = 64

        self.data = <PyObject **>PyMem_Malloc(sizeof(PyObject *) * self.size)

        if self.data == NULL:
            PyErr_NoMemory()

        self.refs = {}

        return 0

    cpdef object getByReference(self, Py_ssize_t ref):
        if ref < 0 or ref >= self.length:
            return None

        return <object>self.data[ref]

    cdef object _ref(self, object obj):
        if self.use_hash:
            return hash(obj)

        return PyLong_FromVoidPtr(<void *>obj)

    cpdef Py_ssize_t getReferenceTo(self, object obj) except -2:
        cdef object p = self.refs.get(self._ref(obj), None)

        if p is None:
            return -1

        return <Py_ssize_t>PyLong_AsLong(<object>p)

    cpdef Py_ssize_t append(self, object obj) except -1:
        self._increase_size()

        cdef object h = self._ref(obj)

        self.refs[h] = <object>self.length
        self.data[self.length] = <PyObject *>obj
        Py_INCREF(obj)

        self.length += 1

        return self.length - 1

    def __iter__(self):
        cdef list x = []
        cdef Py_ssize_t idx

        for idx from 0 <= idx < self.length:
            x.append(<object>self.data[idx])

        return iter(x)

    def __len__(self):
        return self.length

    def __richcmp__(self, object other, int op):
        cdef int equal
        cdef Py_ssize_t i
        # this is necessary because cython does not see the c-space vars of the
        # class for this func
        cdef IndexedCollection s = self

        if PyDict_Check(other) == 1:
            equal = s.refs == other
        elif PyList_Check(other) != 1:
            equal = 0
        else:
            equal = 0

            if PyList_GET_SIZE(other) == s.length:
                equal = 1

                for i from 0 <= i < s.length:
                    if <object>PyList_GET_ITEM(other, i) != <object>s.data[i]:
                        equal = 0

                        break

        if op == 2: # ==
            return equal
        if op == 3: # !=
            return not equal
        raise NotImplementedError

    def __getitem__(self, idx):
        return self.getByReference(idx)

    def __copy__(self):
        cdef IndexedCollection n = IndexedCollection(self.use_hash)

        return n


cdef class ByteStringReferenceCollection(IndexedCollection):
    """
    There have been rare hash collisions within a single AMF payload causing
    corrupt payloads.

    Which strings cause collisions is dependent on the python runtime, each
    platform might have a slightly different implementation which means that
    testing is extremely difficult.
    """

    cdef object _ref(self, object obj):
        return obj

    def __copy__(self):
        cdef ByteStringReferenceCollection n = ByteStringReferenceCollection()

        return n


cdef class Context(object):
    """
    I hold the AMF context for en/decoding streams.

    @ivar objects: An indexed collection of referencable objects encountered
        during en/decoding.
    @type objects: L{util.IndexedCollection}
    @ivar class_aliases: A L{dict} of C{class} to L{ClassAlias}
    """

    def __cinit__(self):
        self.objects = IndexedCollection()
        self.forbid_entities = True
        self.forbid_dtd = True

        self.clear()

    def __init__(self, forbid_dtd=True, forbid_entities=True, **kwargs):
        self.clear()

        self.forbid_entities = forbid_entities
        self.forbid_dtd = forbid_dtd

    cpdef int clear(self) except -1:
        self.objects.clear()

        self.class_aliases = {}
        self.unicodes = {}
        self._strings = {}
        self.extra = {}

        return 0

    cpdef object getObject(self, Py_ssize_t ref):
        return self.objects.getByReference(ref)

    cpdef Py_ssize_t getObjectReference(self, object obj) except -2:
        return self.objects.getReferenceTo(obj)

    cpdef Py_ssize_t addObject(self, object obj) except -1:
        return self.objects.append(obj)

    cpdef object getClassAlias(self, object klass):
        """
        Gets a class alias based on the supplied C{klass}.

        @param klass: The class object.
        @return: The L{ClassAlias} that is linked to C{klass}
        """
        try:
            return self.class_aliases[klass]
        except KeyError:
            pass

        try:
            alias = miniamf.get_class_alias(klass)
        except miniamf.UnknownClassAlias:
            if isinstance(klass, basestring):
                raise

            # no alias has been found yet .. check subclasses
            alias = util.get_class_alias(klass) or miniamf.ClassAlias
            meta = util.get_class_meta(klass)
            alias = alias(klass, defer=True, **meta)

        self.class_aliases[klass] = alias

        return alias

    cpdef unicode getStringForBytes(self, object s):
        """
        Returns the corresponding unicode object for a given string. If there
        is no unicode object, one is created.

        @since: 0.6
        """
        cdef object ret = self.unicodes.get(s, None)

        if ret is not None:
            return ret

        cdef unicode u = s.decode('utf-8')

        self.unicodes[s] = u
        self._strings[u] = s

        return u

    cpdef bytes getBytesForString(self, object u):
        """
        Returns the corresponding utf-8 encoded string for a given unicode
        object. If there is no string, one is encoded.

        @since: 0.6
        """

        cdef object ret = self._strings.get(u, None)

        if ret is not None:
            return ret

        cdef bytes s = u.encode('utf-8')

        self.unicodes[s] = u
        self._strings[u] = s

        return s


cdef class Codec(object):
    """
    Base class for Encoder/Decoder classes. Provides base functionality for
    managing codecs.
    """

    property stream:
        def __get__(self):
            return <BufferedByteStream>self.stream

        def __set__(self, value):
            if not isinstance(value, BufferedByteStream):
                value = BufferedByteStream(value)

            self.stream = <cBufferedByteStream>value

    def __cinit__(self):
        self.stream = None
        self.strict = 0
        self.timezone_offset = None

    def __init__(self, stream=None, strict=False, timezone_offset=None,
                 forbid_entities=True, forbid_dtd=True):
        if not isinstance(stream, BufferedByteStream):
            stream = BufferedByteStream(stream)

        self.stream = <cBufferedByteStream>stream
        self.strict = strict

        self.timezone_offset = timezone_offset

        self.context.forbid_entities = <bint>forbid_entities
        self.context.forbid_dtd = <bint>forbid_dtd


cdef class Decoder(Codec):
    """
    Base AMF decoder.
    """

    def __cinit__(self):
        self.depth = 0

    cdef object readDate(self):
        raise NotImplementedError

    cpdef object readString(self):
        raise NotImplementedError

    cdef object readObject(self):
        raise NotImplementedError

    cdef object readNumber(self):
        raise NotImplementedError

    cdef inline object readNull(self):
        return None

    cdef inline object readUndefined(self):
        return Undefined

    cdef object readList(self):
        raise NotImplementedError

    cdef object readXML(self):
        raise NotImplementedError

    cdef object _readElement(self):
        """
        Reads an element from the data stream.
        """
        cdef Py_ssize_t pos = self.stream.tell()
        cdef char t

        if self.stream.at_eof():
            raise miniamf.EOStream

        t = self.stream.read_char()

        try:
            return self.readConcreteElement(t)
        except IOError:
            self.stream.seek(pos)

            raise

    cpdef object readElement(self):
        cdef object element

        self.depth += 1

        try:
            element = self._readElement()
        finally:
            self.depth -= 1

        if self.depth == 0:
            element = self.finalise(element)

        return element

    cdef object readConcreteElement(self, char t):
        """
        The workhorse function. Overridden in subclasses
        """
        raise NotImplementedError

    cpdef int send(self, data) except -1:
        """
        Add data for the decoder to work on.
        """
        return self.stream.append(data)

    def __next__(self):
        """
        Part of the iterator protocol.
        """
        try:
            return self.readElement()
        except miniamf.EOStream:
            # all data was successfully decoded from the stream
            raise StopIteration

    def __iter__(self):
        return self

    cdef object finalise(self, object payload):
        """
        Finalise the payload.

        This provides a useful hook to adapters to modify the payload that was
        decoded.
        """
        for c in miniamf.POST_DECODE_PROCESSORS:
            payload = c(payload, self.context.extra)

        return payload


cdef class Encoder(Codec):
    """
    Base AMF encoder.
    """

    def __cinit__(self):
        self.func_cache = {}
        self.use_write_object = []
        self.bucket = []

    cpdef int serialiseString(self, u) except -1:
        raise NotImplementedError

    cdef inline int writeType(self, char type) except -1:
        return self.stream.write(<char *>&type, 1)

    cdef int writeNull(self, object o) except -1:
        raise NotImplementedError

    cdef int writeUndefined(self, object o) except -1:
        raise NotImplementedError

    cdef int writeString(self, object o) except -1:
        raise NotImplementedError

    cdef int writeBytes(self, object o) except -1:
        raise NotImplementedError

    cdef int writeBoolean(self, object o) except -1:
        raise NotImplementedError

    cdef int writeInt(self, object o) except -1:
        raise NotImplementedError

    cdef int writeLong(self, object o) except -1:
        raise NotImplementedError

    cdef int writeNumber(self, object o) except -1:
        raise NotImplementedError

    cdef int writeDateTime(self, object o) except -1:
        raise NotImplementedError

    cdef int writeDate(self, object o) except -1:
        o = datetime.datetime.combine(o, datetime.time(0, 0, 0, 0))

        self.writeDateTime(o)

    cdef int writeXML(self, object o) except -1:
        raise NotImplementedError

    cpdef int writeList(self, object o) except -1:
        raise NotImplementedError

    cdef int writeTuple(self, object o) except -1:
        raise NotImplementedError

    cdef int writeDict(self, dict o) except -1:
        raise NotImplementedError

    cdef int writeGenerator(self, object o) except -1:
        while True:
            try:
                self.writeElement(next(o))
            except StopIteration:
                return 0

    cdef int writeSequence(self, object iterable) except -1:
        """
        Encodes an iterable. The default is to write If the iterable has an al
        """
        try:
            alias = self.context.getClassAlias(iterable.__class__)
        except (AttributeError, miniamf.UnknownClassAlias):
            return self.writeList(list(iterable))

        if alias.external:
            # a is a subclassed list with a registered alias - push to the
            # correct method
            self.use_write_object.append(type(iterable))

            return self.writeObject(iterable)

        return self.writeList(list(iterable))

    cpdef int writeObject(self, object o) except -1:
        raise NotImplementedError

    cdef int writeMixedArray(self, object o) except -1:
        raise NotImplementedError

    cdef int handleBasicTypes(self, object element, object py_type) except -1:
        """
        @return: 0 = handled, -1 = error, 1 = not handled
        """
        if PyBytes_Check(element):
            return self.writeBytes(element)
        elif PyUnicode_Check(element):
            return self.writeString(element)
        elif element is None:
            return self.writeNull(element)
        elif PyBool_Check(element):
            return self.writeBoolean(element)
        elif PyLong_CheckExact(element):
            return self.writeLong(element)
        elif PyFloat_CheckExact(element):
            return self.writeNumber(element)
        elif PyList_CheckExact(element):
            return self.writeList(element)
        elif PyTuple_CheckExact(element):
            return self.writeTuple(element)
        elif element is Undefined:
            return self.writeUndefined(element)
        elif PyDict_CheckExact(element):
            return self.writeDict(element)
        elif PyDateTime_CheckExact(element):
            return self.writeDateTime(element)
        elif PyDate_CheckExact(element):
            return self.writeDate(element)
        elif py_type is MixedArray:
            return self.writeMixedArray(element)
        elif py_type is GeneratorType:
            return self.writeGenerator(element)
        elif PySequence_Contains(self.use_write_object, py_type):
            return self.writeObject(element)
        elif xml.is_xml(element):
            return self.writeXML(element)
        return 1

    cdef int checkBadTypes(self, object element, object py_type) except -1:
        if PyModule_CheckExact(element):
            raise miniamf.EncodeError("Cannot encode modules %r" % (element,))
        elif PyMethod_Check(element):
            raise miniamf.EncodeError("Cannot encode methods %r" % (element,))
        elif PyFunction_Check(element) or py_type is BuiltinFunctionType:
            raise miniamf.EncodeError("Cannot encode functions %r" % (
                element,
            ))
        elif PyType_Check(element) or PyType_CheckExact(element):
            raise miniamf.EncodeError("Cannot encode class objects %r" % (
                element,
            ))
        elif PyTime_CheckExact(element):
            raise miniamf.EncodeError('A datetime.time instance was found but '
                'AMF has no way to encode time objects. Please use '
                'datetime.datetime instead')

        return 0

    cpdef int writeElement(self, object element) except -1:
        cdef int ret = 0
        cdef object py_type = type(element)
        cdef object func = None

        ret = self.handleBasicTypes(element, py_type)

        if ret == 1:
            # encoding was not handled by basic types
            func = self.func_cache.get(py_type, None)

            if func is None:
                func = get_custom_type_func(self, element)

                if func is None:
                    self.checkBadTypes(element, py_type)
                    self.use_write_object.append(py_type)

                    if isinstance(element, (list, tuple)):
                        return self.writeSequence(element)

                    return self.writeObject(element)

                self.func_cache[py_type] = func

            func(element)

        return ret

    cpdef int send(self, data) except -1:
        """
        Add data for the decoder to work on.
        """
        self.bucket.append(data)

    def __next__(self):
        """
        Part of the iterator protocol.
        """
        cdef Py_ssize_t start_pos, end_pos
        cdef char *buf = NULL

        try:
            element = self.bucket.pop(0)
        except IndexError:
            raise StopIteration

        start_pos = self.stream.tell()

        self.writeElement(element)

        end_pos = self.stream.tell()

        self.stream.seek(start_pos)

        self.stream.read(&buf, end_pos - start_pos)

        return PyBytes_FromStringAndSize(buf, end_pos - start_pos)

    def __iter__(self):
        return self


cdef class _CustomTypeFunc(object):
    """
    Support for custom type mappings when encoding.
    """

    cdef Encoder encoder
    cdef object func

    def __cinit__(self, Encoder encoder, func):
        self.encoder = encoder
        self.func = func

    def __call__(self, data, **kwargs):
        ret = self.func(data, encoder=self.encoder)

        if ret is not None:
            self.encoder.writeElement(ret)


cdef object get_custom_type_func(object encoder, object data):
    cdef _CustomTypeFunc ret

    for type_, func in miniamf.TYPE_MAP.items():
        try:
            if isinstance(data, type_):
                return _CustomTypeFunc(encoder, func)
        except TypeError:
            if callable(type_) and type_(data):
                return _CustomTypeFunc(encoder, func)

    return None



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Tests for the L{collections} L{miniamf.adapters._collections} module.

@since: 0.5
"""

import collections
import unittest

import miniamf


class CollectionsTestCase(unittest.TestCase):
    """
    """

    def encdec(self, encoding):
        return next(miniamf.decode(
            miniamf.encode(self.obj, encoding=encoding),
            encoding=encoding))


class DequeTestCase(CollectionsTestCase):
    """
    Tests for L{collections.deque}
    """

    def setUp(self):
        CollectionsTestCase.setUp(self)

        self.orig = [1, 2, 3]
        self.obj = collections.deque(self.orig)

    def test_amf0(self):
        self.assertEqual(self.encdec(miniamf.AMF0), self.orig)

    def test_amf3(self):
        self.assertEqual(self.encdec(miniamf.AMF3), self.orig)


class DefaultDictTestCase(CollectionsTestCase):
    """
    Tests for L{collections.defaultdict}
    """

    def setUp(self):
        CollectionsTestCase.setUp(self)

        s = 'mississippi'
        self.obj = collections.defaultdict(int)

        for k in s:
            self.obj[k] += 1

        self.orig = dict(self.obj)

    def test_amf0(self):
        self.assertEqual(self.encdec(miniamf.AMF0), self.orig)

    def test_amf3(self):
        self.assertEqual(self.encdec(miniamf.AMF3), self.orig)


class OrderedDictTestCase(CollectionsTestCase):
    """
    Tests for L{collections.OrderedDict}
    """

    def setUp(self):
        CollectionsTestCase.setUp(self)

        self.obj = collections.OrderedDict(
            [('apple', 4), ('banana', 3), ('orange', 2), ('pear', 1)]
        )
        self.orig = dict(self.obj)

    def test_amf0(self):
        self.assertEqual(self.encdec(miniamf.AMF0), self.orig)

    def test_amf3(self):
        self.assertEqual(self.encdec(miniamf.AMF3), self.orig)


class CounterTestCase(CollectionsTestCase):
    """
    Tests for L{collections.Counter}
    """

    def setUp(self):
        CollectionsTestCase.setUp(self)

        self.obj = collections.Counter({'blue': 3, 'red': 2, 'green': 1})

        self.orig = dict(self.obj)

    def test_amf0(self):
        self.assertEqual(self.encdec(miniamf.AMF0), self.orig)

    def test_amf3(self):
        self.assertEqual(self.encdec(miniamf.AMF3), self.orig)


class NamedTupleTestCase(CollectionsTestCase):
    """
    Tests for L{collections.namedtuple}
    """

    def setUp(self):
        CollectionsTestCase.setUp(self)

        user_vo = collections.namedtuple('user_vo', 'id name age')

        miniamf.add_type(user_vo, lambda obj, encoder: obj._asdict())

        self.obj = user_vo(1, 'Hadrien', 30)
        self.orig = self.obj._asdict()

    def test_amf0(self):
        self.assertEqual(self.encdec(miniamf.AMF0), self.orig)

    def test_amf3(self):
        self.assertEqual(self.encdec(miniamf.AMF3), self.orig)



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
Class alias base functionality.

@since: 0.6
"""

import inspect

import miniamf
from . import util


class UnknownClassAlias(Exception):
    """
    Raised if the AMF stream specifies an Actionscript class that does not
    have a Python class alias.

    @see: L{register_class}
    """


class ClassAlias(object):
    """
    Class alias. Provides class/instance meta data to the En/Decoder to allow
    fine grain control and some performance increases.
    """

    def __init__(self, klass, alias=None, **kwargs):
        if not isinstance(klass, type):
            raise TypeError("klass must be a class type, got %r" % klass)

        # In Python 2, both byte and Unicode string types have .decode
        # methods.  In Python 3, only byte strings do.
        if alias is None or isinstance(alias, str):
            pass
        elif hasattr(alias, "decode"):
            alias = alias.decode("utf-8")

        self.checkClass(klass)

        self.klass = klass
        self.alias = alias

        self.static_attrs = kwargs.pop('static_attrs', None)
        self.exclude_attrs = kwargs.pop('exclude_attrs', None)
        self.readonly_attrs = kwargs.pop('readonly_attrs', None)
        self.amf3 = kwargs.pop('amf3', None)
        self.external = kwargs.pop('external', None)
        self.dynamic = kwargs.pop('dynamic', None)
        self.synonym_attrs = kwargs.pop('synonym_attrs', {})

        self._compiled = False
        self.anonymous = False
        self.sealed = None
        self.bases = None

        if self.alias is None:
            self.anonymous = True
            # we don't set this to None because AMF3 untyped objects have a
            # class name of ''
            self.alias = ''
        else:
            if self.alias == '':
                raise ValueError('Cannot set class alias as \'\'')

        if not kwargs.pop('defer', False):
            self.compile()

        if kwargs:
            raise TypeError('Unexpected keyword arguments %r' % (kwargs,))

    def _checkExternal(self):
        k = self.klass

        if not hasattr(k, '__readamf__'):
            raise AttributeError(
                "An externalised class was specified, but"
                " no __readamf__ attribute was found for %r" % (k,)
            )

        if not hasattr(k, '__writeamf__'):
            raise AttributeError(
                "An externalised class was specified, but"
                " no __writeamf__ attribute was found for %r" % (k,)
            )

        if not hasattr(k.__readamf__, '__call__'):
            raise TypeError("%s.__readamf__ must be callable" % (k.__name__,))

        if not hasattr(k.__writeamf__, '__call__'):
            raise TypeError("%s.__writeamf__ must be callable" % (k.__name__,))

    def compile(self):
        """
        This compiles the alias into a form that can be of most benefit to the
        en/decoder.
        """
        if self._compiled:
            return

        self.decodable_properties = set()
        self.encodable_properties = set()
        self.inherited_dynamic = None
        self.inherited_sealed = None
        self.bases = []

        self.exclude_attrs = set(self.exclude_attrs or [])
        self.readonly_attrs = set(self.readonly_attrs or [])
        self.static_attrs = list(self.static_attrs or [])
        self.static_attrs_set = set(self.static_attrs)

        self.sealed = util.is_class_sealed(self.klass)

        if self.external:
            self._checkExternal()
            self._finalise_compile()

            # this class is external so no more compiling is necessary
            return

        if hasattr(self.klass, '__slots__'):
            self.decodable_properties.update(self.klass.__slots__)
            self.encodable_properties.update(self.klass.__slots__)

        for k, v in self.klass.__dict__.items():
            if not isinstance(v, property):
                continue

            if v.fget:
                self.encodable_properties.update([k])

            if v.fset:
                self.decodable_properties.update([k])
            else:
                self.readonly_attrs.update([k])

        mro = inspect.getmro(self.klass)[1:]

        for c in mro:
            self._compile_base_class(c)

        self.getCustomProperties()

        self._finalise_compile()

    def _compile_base_class(self, klass):
        if klass is object:
            return

        try:
            alias = miniamf.get_class_alias(klass)
        except UnknownClassAlias:
            alias = miniamf.register_class(klass)

        alias.compile()

        self.bases.append((klass, alias))

        if alias.exclude_attrs:
            self.exclude_attrs.update(alias.exclude_attrs)

        if alias.readonly_attrs:
            self.readonly_attrs.update(alias.readonly_attrs)

        if alias.static_attrs:
            self.static_attrs_set.update(alias.static_attrs)

            for a in alias.static_attrs:
                if a not in self.static_attrs:
                    self.static_attrs.insert(0, a)

        if alias.encodable_properties:
            self.encodable_properties.update(alias.encodable_properties)

        if alias.decodable_properties:
            self.decodable_properties.update(alias.decodable_properties)

        if self.amf3 is None and alias.amf3:
            self.amf3 = alias.amf3

        if self.dynamic is None and alias.dynamic is not None:
            self.inherited_dynamic = alias.dynamic

        if alias.sealed is not None:
            self.inherited_sealed = alias.sealed

        if alias.synonym_attrs:
            x = self.synonym_attrs
            self.synonym_attrs = alias.synonym_attrs.copy()
            self.synonym_attrs.update(x)

    def _finalise_compile(self):
        if self.dynamic is None:
            self.dynamic = True

            if self.inherited_dynamic is not None:
                if not self.inherited_dynamic \
                        and not self.sealed \
                        and self.inherited_sealed:
                    self.dynamic = True
                else:
                    self.dynamic = self.inherited_dynamic

        if self.sealed:
            self.dynamic = False

        if self.amf3 is None:
            self.amf3 = False

        if self.external is None:
            self.external = False

        if self.static_attrs:
            self.static_attrs = list(self.static_attrs)
            self.static_attrs.sort()

            self.encodable_properties.update(self.static_attrs)
            self.decodable_properties.update(self.static_attrs)

        if self.static_attrs:
            self.static_attrs_set.update(self.static_attrs)

            if self.exclude_attrs:
                self.static_attrs_set.difference_update(self.exclude_attrs)

            for a in self.static_attrs_set:
                if a not in self.static_attrs:
                    self.static_attrs.remove(a)

        if not self.exclude_attrs:
            self.exclude_attrs = None
        else:
            self.encodable_properties.difference_update(self.exclude_attrs)
            self.decodable_properties.difference_update(self.exclude_attrs)

        if self.exclude_attrs is not None:
            self.exclude_attrs = list(self.exclude_attrs)
            self.exclude_attrs.sort()

        if not self.readonly_attrs:
            self.readonly_attrs = None
        else:
            self.decodable_properties.difference_update(self.readonly_attrs)

        if self.readonly_attrs is not None:
            self.readonly_attrs = list(self.readonly_attrs)
            self.readonly_attrs.sort()

        if len(self.decodable_properties) == 0:
            self.decodable_properties = None
        else:
            self.decodable_properties = list(self.decodable_properties)
            self.decodable_properties.sort()

        if len(self.encodable_properties) == 0:
            self.encodable_properties = None
        else:
            self.encodable_properties = list(self.encodable_properties)
            self.encodable_properties.sort()

        self.non_static_encodable_properties = None

        if self.encodable_properties:
            self.non_static_encodable_properties = set(
                self.encodable_properties
            )

            if self.static_attrs:
                self.non_static_encodable_properties.difference_update(
                    self.static_attrs
                )

        self.shortcut_encode = True
        self.shortcut_decode = True

        if (self.encodable_properties or self.static_attrs or
                self.exclude_attrs or self.external or
                self.synonym_attrs):
            self.shortcut_encode = False

        if (self.decodable_properties or self.static_attrs or
                self.exclude_attrs or self.readonly_attrs or
                not self.dynamic or self.external or self.synonym_attrs):
            self.shortcut_decode = False

        self.is_dict = False

        if issubclass(self.klass, dict) or self.klass is dict:
            self.is_dict = True

        self._compiled = True

    def is_compiled(self):
        return self._compiled

    def __str__(self):
        return self.alias

    def __bytes__(self):
        return self.alias.encode('utf-8')

    def __repr__(self):
        k = self.__class__

        return '<%s.%s alias=%r class=%r @ 0x%x>' % (
            k.__module__,
            k.__name__,
            self.alias,
            self.klass,
            id(self)
        )

    def __eq__(self, other):
        if isinstance(other, str):
            return self.alias == other
        elif isinstance(other, self.__class__):
            return self.klass == other.klass
        elif isinstance(other, type):
            return self.klass == other
        else:
            return False

    def __hash__(self):
        return id(self)

    def checkClass(self, klass):
        """
        This function is used to check if the class being aliased fits certain
        criteria. The default is to check that C{__new__} is available or the
        C{__init__} constructor does not need additional arguments. If this is
        the case then L{TypeError} will be raised.

        @since: 0.4
        """
        # Check for __new__ support.
        if hasattr(klass, '__new__') and hasattr(klass.__new__, '__call__'):
            # Should be good to go.
            return

        # Check that the constructor of the class doesn't require any additonal
        # arguments.
        if not (
            hasattr(klass, '__init__') and hasattr(klass.__init__, '__call__')
        ):
            return

        klass_func = klass.__init__.__func__

        if not hasattr(klass_func, 'func_code'):
            # Can't examine it, assume it's OK.
            return

        if klass_func.__defaults__:
            available_arguments = len(klass_func.__defaults__) + 1
        else:
            available_arguments = 1

        needed_arguments = klass_func.__code__.co_argcount

        if available_arguments >= needed_arguments:
            # Looks good to me.
            return

        spec = inspect.getargspec(klass_func)

        raise TypeError(
            "__init__ doesn't support additional arguments: %s" % (
                inspect.formatargspec(*spec)
            )
        )

    def getAttribute(self, obj, attr, codec=None):
        """
        Get the attribute `attr` from `obj`. If no attribute exists,
        `miniamf.Undefined` is returned.

        @param codec: The current `miniamf.codec.Codec` getting the attribute
            (if there is one).
        """
        return getattr(obj, attr)

    def getEncodableAttributes(self, obj, codec=None):
        """
        Must return a C{dict} of attributes to be encoded, even if its empty.

        @param codec: An optional argument that will contain the encoder
            instance calling this function.
        @since: 0.5
        """
        if not self._compiled:
            self.compile()

        if self.is_dict:
            return dict(obj)

        if self.shortcut_encode and self.dynamic:
            return obj.__dict__.copy()

        attrs = {}

        if self.static_attrs:
            for attr in self.static_attrs:
                try:
                    attrs[attr] = self.getAttribute(obj, attr, codec=codec)
                except AttributeError:
                    attrs[attr] = miniamf.Undefined

        if not self.dynamic:
            if self.non_static_encodable_properties:
                for attr in self.non_static_encodable_properties:
                    attrs[attr] = self.getAttribute(obj, attr, codec=codec)

            return attrs

        dynamic_props = util.get_properties(obj)

        if not self.shortcut_encode:
            dynamic_props = set(dynamic_props)

            if self.encodable_properties:
                dynamic_props.update(self.encodable_properties)

            if self.static_attrs:
                dynamic_props.difference_update(self.static_attrs)

            if self.exclude_attrs:
                dynamic_props.difference_update(self.exclude_attrs)

        for attr in dynamic_props:
            attrs[attr] = self.getAttribute(obj, attr, codec=codec)

        if self.synonym_attrs:
            missing = object()

            for k, v in self.synonym_attrs.items():
                value = attrs.pop(k, missing)

                if value is missing:
                    continue

                attrs[v] = value

        return attrs

    def getDecodableAttributes(self, obj, attrs, codec=None):
        """
        Returns a dictionary of attributes for C{obj} that has been filtered,
        based on the supplied C{attrs}. This allows for fine grain control
        over what will finally end up on the object or not.

        @param obj: The object that will recieve the attributes.
        @param attrs: The C{attrs} dictionary that has been decoded.
        @param codec: An optional argument that will contain the decoder
            instance calling this function.
        @return: A dictionary of attributes that can be applied to C{obj}
        @since: 0.5
        """
        if not self._compiled:
            self.compile()

        changed = False

        props = set(attrs.keys())

        if self.static_attrs:
            missing_attrs = self.static_attrs_set.difference(props)

            if missing_attrs:
                raise AttributeError(
                    'Static attributes %r expected when decoding %r' % (
                        missing_attrs, self.klass
                    )
                )

            props.difference_update(self.static_attrs)

        if not props:
            return attrs

        if not self.dynamic:
            if not self.decodable_properties:
                props = set()
            else:
                props.intersection_update(self.decodable_properties)

            changed = True

        if self.readonly_attrs:
            props.difference_update(self.readonly_attrs)
            changed = True

        if self.exclude_attrs:
            props.difference_update(self.exclude_attrs)
            changed = True

        if changed:
            # apply all filters before synonyms
            a = {}

            [a.__setitem__(p, attrs[p]) for p in props]
            attrs = a

        if self.synonym_attrs:
            missing = object()

            for k, v in self.synonym_attrs.items():
                value = attrs.pop(v, missing)

                if value is missing:
                    continue

                attrs[k] = value

        return attrs

    def applyAttributes(self, obj, attrs, codec=None):
        """
        Applies the collection of attributes C{attrs} to aliased object C{obj}.
        Called when decoding reading aliased objects from an AMF byte stream.

        Override this to provide fine grain control of application of
        attributes to C{obj}.

        @param codec: An optional argument that will contain the en/decoder
            instance calling this function.
        """
        if not self._compiled:
            self.compile()

        if not self.shortcut_decode:
            attrs = self.getDecodableAttributes(obj, attrs, codec=codec)
        else:
            if self.is_dict:
                obj.update(attrs)

                return

            if not self.sealed:
                obj.__dict__.update(attrs)

                return

        util.set_attrs(obj, attrs)

    def getCustomProperties(self):
        """
        Overrride this to provide known static properties based on the aliased
        class.

        @since: 0.5
        """

    def createInstance(self, codec=None):
        """
        Creates an instance of the klass.

        @return: Instance of C{self.klass}.
        """
        if type(self.klass) is type:
            return self.klass.__new__(self.klass)

        return self.klass()



# Copyright (c) The PyAMF Project.
# See LICENSE.txt for details.

"""
U{array<http://docs.python.org/library/array.html>} adapter module.

Will convert all array.array instances to a python list before encoding. All
type information is lost (but degrades nicely).

@since: 0.5
"""

import array

import miniamf
from miniamf.adapters import util

miniamf.add_type(array.ArrayType, util.to_list)
